import re
import sys

from ..test_managercli import TestCliCommand
from subscription_manager import managercli
from subscription_manager.repolib import Repo

from ..fixture import Capture, Matcher, set_up_mock_sp_store

from unittest.mock import patch, Mock, call


class TestReposCommand(TestCliCommand):
    command_class = managercli.ReposCommand

    def setUp(self):
        super(TestReposCommand, self).setUp(False)
        argv_patcher = patch.object(sys, "argv", ["subscription-manager", "repos"])
        argv_patcher.start()
        self.addCleanup(argv_patcher.stop)
        self.cc.cp = Mock()
        syspurpose_patch = patch("subscription_manager.syspurposelib.SyncedStore")
        self.mock_sp_store = syspurpose_patch.start()
        self.mock_sp_store, self.mock_sp_store_contents = set_up_mock_sp_store(self.mock_sp_store)
        self.addCleanup(syspurpose_patch.stop)

    def check_output_for_repos(self, output, repos):
        """
        Checks the given output string for the specified repos' ids.

        Returns a tuple of booleans specifying whether or not the repo in the corresponding position
        was found in the output.
        """
        searches = []
        for repo in repos:
            # Impl note: This may break if a repo's ID contains special regex characters.
            searches.append(re.search("^Repo ID:\\s+%s$" % repo.id, output, re.MULTILINE) is not None)

        return tuple(searches)

    def check_output_for_repos_enabled_disabled(self, output, repos):
        """
        Checks that the given output string has the sequence of
        "Repository X is enabled/disabled" lines matching the requested repos.
        """
        lines_set = set(output.splitlines())
        repos_set = set()
        for (repo_name, repo_enabled) in repos:
            if repo_enabled:
                repos_set.add(f"Repository '{repo_name}' is enabled for this system.")
            else:
                repos_set.add(f"Repository '{repo_name}' is disabled for this system.")
        self.assertEqual(repos_set, lines_set)

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_default(self, mock_invoker):
        self.cc.main()
        self.cc._validate_options()

        repos = [Repo("x", [("enabled", "1")]), Repo("y", [("enabled", "0")]), Repo("z", [("enabled", "0")])]
        mock_invoker.return_value.get_repos.return_value = repos

        with Capture() as cap:
            self.cc._do_command()

        result = self.check_output_for_repos(cap.out, repos)
        self.assertEqual((True, True, True), result)

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_list(self, mock_invoker):
        self.cc.main(["--list"])
        self.cc._validate_options()

        repos = [Repo("x", [("enabled", "1")]), Repo("y", [("enabled", "0")]), Repo("z", [("enabled", "0")])]
        mock_invoker.return_value.get_repos.return_value = repos

        with Capture() as cap:
            self.cc._do_command()

        result = self.check_output_for_repos(cap.out, repos)
        self.assertEqual((True, True, True), result)

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_list_with_enabled(self, mock_invoker):
        self.cc.main(["--list", "--list-enabled"])
        self.cc._validate_options()

        repos = [Repo("x", [("enabled", "1")]), Repo("y", [("enabled", "0")]), Repo("z", [("enabled", "0")])]
        mock_invoker.return_value.get_repos.return_value = repos

        with Capture() as cap:
            self.cc._do_command()

        result = self.check_output_for_repos(cap.out, repos)
        self.assertEqual((True, True, True), result)

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_list_with_disabled(self, mock_invoker):
        self.cc.main(["--list", "--list-disabled"])
        self.cc._validate_options()

        repos = [Repo("x", [("enabled", "1")]), Repo("y", [("enabled", "0")]), Repo("z", [("enabled", "0")])]
        mock_invoker.return_value.get_repos.return_value = repos

        with Capture() as cap:
            self.cc._do_command()

        result = self.check_output_for_repos(cap.out, repos)
        self.assertEqual((True, True, True), result)

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_list_with_enabled_and_disabled(self, mock_invoker):
        self.cc.main(["--list", "--list-disabled", "--list-disabled"])
        self.cc._validate_options()

        repos = [Repo("x", [("enabled", "1")]), Repo("y", [("enabled", "0")]), Repo("z", [("enabled", "0")])]
        mock_invoker.return_value.get_repos.return_value = repos

        with Capture() as cap:
            self.cc._do_command()

        result = self.check_output_for_repos(cap.out, repos)
        self.assertEqual((True, True, True), result)

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_list_enabled(self, mock_invoker):
        self.cc.main(["--list-enabled"])
        self.cc._validate_options()

        repos = [
            Repo("x", [("enabled", "1")]),
            Repo("y", [("enabled", "0")]),
            Repo("z", [("enabled", "0")]),
            Repo("a", [("enabled", "false")]),
            Repo("b", [("enabled", "False")]),
            Repo("c", [("enabled", "true")]),
        ]
        mock_invoker.return_value.get_repos.return_value = repos

        with Capture() as cap:
            self.cc._do_command()

        result = self.check_output_for_repos(cap.out, repos)
        self.assertEqual((True, False, False, False, False, True), result)

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_list_disabled(self, mock_invoker):
        self.cc.main(["--list-disabled"])
        self.cc._validate_options()

        repos = [Repo("x", [("enabled", "1")]), Repo("y", [("enabled", "0")]), Repo("z", [("enabled", "0")])]
        mock_invoker.return_value.get_repos.return_value = repos

        with Capture() as cap:
            self.cc._do_command()

        result = self.check_output_for_repos(cap.out, repos)
        self.assertEqual((False, True, True), result)

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_list_enabled_and_disabled(self, mock_invoker):
        self.cc.main(["--list-enabled", "--list-disabled"])
        self.cc._validate_options()

        repos = [Repo("x", [("enabled", "1")]), Repo("y", [("enabled", "0")]), Repo("z", [("enabled", "0")])]
        mock_invoker.return_value.get_repos.return_value = repos

        with Capture() as cap:
            self.cc._do_command()

        result = self.check_output_for_repos(cap.out, repos)
        self.assertEqual((True, True, True), result)

    def test_enable(self):
        self.cc.main(["--enable", "one", "--enable", "two"])
        self.cc._validate_options()

    def test_disable(self):
        self.cc.main(["--disable", "one", "--disable", "two"])
        self.cc._validate_options()

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_list_enable_and_disable(self, mock_invoker):
        self._inject_mock_invalid_consumer()
        self.cc.main(["--disable", "x", "--enable", "x", "--enable", "z", "--disable", "z"])
        self.cc._validate_options()

        repos = [Repo("x", [("enabled", "1")]), Repo("y", [("enabled", "0")]), Repo("z", [("enabled", "0")])]
        mock_invoker.return_value.get_repos.return_value = repos

        with Capture() as cap:
            self.cc._do_command()

        repos_changes = [("x", True), ("z", False)]
        self.check_output_for_repos_enabled_disabled(cap.out, repos_changes)

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_list_enable_and_disable_wildcards(self, mock_invoker):
        self._inject_mock_invalid_consumer()
        self.cc.main(["--disable", "*", "--enable", "z"])
        self.cc._validate_options()

        repos = [Repo("x", [("enabled", "1")]), Repo("y", [("enabled", "1")]), Repo("z", [("enabled", "1")])]
        mock_invoker.return_value.get_repos.return_value = repos

        with Capture() as cap:
            self.cc._do_command()

        repos_changes = [("x", False), ("y", False), ("z", True)]
        self.check_output_for_repos_enabled_disabled(cap.out, repos_changes)

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_set_repo_status(self, mock_repolib):
        repolib_instance = mock_repolib.return_value
        self._inject_mock_valid_consumer("fake_id")

        repos = [Repo("x"), Repo("y"), Repo("z")]
        items = [("0", "x"), ("0", "y")]
        self.cc.use_overrides = True
        self.cc._set_repo_status(repos, repolib_instance, items)

        expected_overrides = [{"contentLabel": i, "name": "enabled", "value": "0"} for (_action, i) in items]
        metadata_overrides = [
            {"contentLabel": i, "name": "enabled_metadata", "value": "0"} for (_action, i) in items
        ]
        expected_overrides.extend(metadata_overrides)

        # The list of overrides sent to setContentOverrides is really a set of
        # dictionaries (since we don't know the order of the overrides).
        # However, since the dict class is not hashable, we can't actually use
        # a set.  So we need a custom matcher to make sure that the
        # JSON passed in to setContentOverrides is what we expect.
        match_dict_list = Matcher(self.assert_items_equals, expected_overrides)
        self.cc.cp.setContentOverrides.assert_called_once_with("fake_id", match_dict_list)
        self.assertTrue(repolib_instance.update.called)

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_set_repo_status_with_wildcards(self, mock_repolib):
        repolib_instance = mock_repolib.return_value
        self._inject_mock_valid_consumer("fake_id")

        repos = [Repo("zoo"), Repo("zebra"), Repo("zip")]
        items = [("0", "z*")]
        self.cc.use_overrides = True
        self.cc._set_repo_status(repos, repolib_instance, items)

        expected_overrides = [{"contentLabel": i.id, "name": "enabled", "value": "0"} for i in repos]
        metadata_overrides = [{"contentLabel": i.id, "name": "enabled_metadata", "value": "0"} for i in repos]
        expected_overrides.extend(metadata_overrides)
        match_dict_list = Matcher(self.assert_items_equals, expected_overrides)
        self.cc.cp.setContentOverrides.assert_called_once_with("fake_id", match_dict_list)
        self.assertTrue(repolib_instance.update.called)

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_set_repo_status_disable_all_enable_some(self, mock_repolib):
        repolib_instance = mock_repolib.return_value
        self._inject_mock_valid_consumer("fake_id")

        repos = [Repo("zoo"), Repo("zebra"), Repo("zip")]
        items = [("0", "*"), ("1", "zoo"), ("1", "zip")]
        self.cc.use_overrides = True
        self.cc._set_repo_status(repos, repolib_instance, items)

        expected_overrides = [
            {"contentLabel": "zebra", "name": "enabled", "value": "0"},
            {"contentLabel": "zebra", "name": "enabled_metadata", "value": "0"},
            {"contentLabel": "zoo", "name": "enabled", "value": "1"},
            {"contentLabel": "zoo", "name": "enabled_metadata", "value": "1"},
            {"contentLabel": "zip", "name": "enabled", "value": "1"},
            {"contentLabel": "zip", "name": "enabled_metadata", "value": "1"},
        ]
        match_dict_list = Matcher(self.assert_items_equals, expected_overrides)
        self.cc.cp.setContentOverrides.assert_called_once_with("fake_id", match_dict_list)
        self.assertTrue(repolib_instance.update.called)

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_set_repo_status_enable_all_disable_some(self, mock_repolib):
        repolib_instance = mock_repolib.return_value
        self._inject_mock_valid_consumer("fake_id")

        repos = [Repo("zoo"), Repo("zebra"), Repo("zip")]
        items = [("1", "*"), ("0", "zoo"), ("0", "zip")]
        self.cc.use_overrides = True
        self.cc._set_repo_status(repos, repolib_instance, items)

        expected_overrides = [
            {"contentLabel": "zebra", "name": "enabled", "value": "1"},
            {"contentLabel": "zebra", "name": "enabled_metadata", "value": "1"},
            {"contentLabel": "zoo", "name": "enabled", "value": "0"},
            {"contentLabel": "zoo", "name": "enabled_metadata", "value": "0"},
            {"contentLabel": "zip", "name": "enabled", "value": "0"},
            {"contentLabel": "zip", "name": "enabled_metadata", "value": "0"},
        ]
        match_dict_list = Matcher(self.assert_items_equals, expected_overrides)
        self.cc.cp.setContentOverrides.assert_called_once_with("fake_id", match_dict_list)
        self.assertTrue(repolib_instance.update.called)

    @patch("subscription_manager.cli_command.repos.RepoActionInvoker")
    def test_set_repo_status_enable_all_disable_all(self, mock_repolib):
        repolib_instance = mock_repolib.return_value
        self._inject_mock_valid_consumer("fake_id")

        repos = [Repo("zoo"), Repo("zebra"), Repo("zip")]
        items = [("1", "*"), ("0", "*")]
        self.cc.use_overrides = True
        self.cc._set_repo_status(repos, repolib_instance, items)

        expected_overrides = [
            {"contentLabel": "zebra", "name": "enabled", "value": "0"},
            {"contentLabel": "zebra", "name": "enabled_metadata", "value": "0"},
            {"contentLabel": "zoo", "name": "enabled", "value": "0"},
            {"contentLabel": "zoo", "name": "enabled_metadata", "value": "0"},
            {"contentLabel": "zip", "name": "enabled", "value": "0"},
            {"contentLabel": "zip", "name": "enabled_metadata", "value": "0"},
        ]
        match_dict_list = Matcher(self.assert_items_equals, expected_overrides)
        self.cc.cp.setContentOverrides.assert_called_once_with("fake_id", match_dict_list)
        self.assertTrue(repolib_instance.update.called)

    @patch("subscription_manager.repofile.RepoFileBase.path_exists")
    @patch("subscription_manager.cli_command.repos.YumRepoFile")
    def test_set_repo_status_when_disconnected(self, mock_repofile, mock_path_exists):
        mock_path_exists.return_value = True
        self._inject_mock_invalid_consumer()
        mock_repofile_inst = mock_repofile.return_value

        enabled = list({"enabled": "1"}.items())
        disabled = list({"enabled": "0"}.items())

        zoo = Repo("zoo", enabled)
        zebra = Repo("zebra", disabled)
        zippy = Repo("zippy", enabled)
        zero = Repo("zero", disabled)
        repos = [zoo, zebra, zippy, zero]
        items = [("0", "z*")]

        self.cc._set_repo_status(repos, None, items)
        calls = [call(r) for r in repos if r["enabled"] == 1]
        mock_repofile_inst.update.assert_has_calls(calls)
        for r in repos:
            self.assertEqual("0", r["enabled"])
        mock_repofile_inst.write.assert_called_once_with()



#
# Subscription manager command line utility.
#
# Copyright (c) 2021 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#
import logging
import os

from subscription_manager import managerlib, rhelentbranding
from subscription_manager.i18n import ugettext as _
from subscription_manager.cli_command.cli import CliCommand
from subscription_manager.cli import system_exit

log = logging.getLogger(__name__)


class ImportCertCommand(CliCommand):
    def __init__(self):
        shortdesc = _("Import certificates which were provided outside of the tool")
        super(ImportCertCommand, self).__init__("import", shortdesc, False)

        self.parser.add_argument(
            "--certificate",
            action="append",
            dest="certificate_file",
            help=_("certificate file to import (can be specified more than once)"),
        )

    def _validate_options(self):
        if self.is_registered():
            system_exit(
                os.EX_USAGE,
                _(
                    "Error: You may not import certificates into a system that "
                    "is registered to a subscription management service."
                ),
            )
        if not self.options.certificate_file:
            system_exit(
                os.EX_USAGE,
                _("Error: This command requires that you specify a certificate with --certificate."),
            )

    def _do_command(self):
        self._validate_options()
        # Return code
        imported_certs = []
        for src_cert_file in self.options.certificate_file:
            src_cert_file = os.path.expanduser(src_cert_file)
            if os.path.exists(src_cert_file):
                try:
                    extractor = managerlib.ImportFileExtractor(src_cert_file)

                    # Verify the entitlement data.
                    if extractor.verify_valid_entitlement():
                        extractor.write_to_disk()
                        print(
                            _("Successfully imported certificate {file}").format(
                                file=os.path.basename(src_cert_file)
                            )
                        )
                        imported_certs.append(extractor.get_cert())
                    else:
                        log.error(
                            "Error parsing manually imported entitlement "
                            "certificate: {src_cert_file}".format(src_cert_file=src_cert_file)
                        )
                        print(
                            _(
                                "{file} is not a valid certificate file. Please use a valid certificate."
                            ).format(file=os.path.basename(src_cert_file))
                        )

                except Exception as e:
                    # Should not get here unless something really bad happened.
                    log.exception(e)
                    print(
                        _(
                            "An error occurred while importing the certificate. "
                            "Please check log file for more information."
                        )
                    )
            else:
                log.error("Supplied certificate file does not exist: {file}".format(file=src_cert_file))
                print(_("{file}: file not found.").format(file=os.path.basename(src_cert_file)))

        # update branding info for the imported certs, if needed
        if imported_certs:
            # RHELBrandsInstaller will load ent dir by default
            brands_installer = rhelentbranding.RHELBrandsInstaller()
            brands_installer.install()

        self._request_validity_check()

        return_code = 0
        if not imported_certs:
            return_code = 1

        return return_code

    def require_connection(self):
        return False



#
# Subscription manager command line utility.
#
# Copyright (c) 2021 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#
import fileinput
import logging
import os
import re

import rhsm.connection as connection
import subscription_manager.injection as inj

from rhsmlib.services import attach, products

from subscription_manager.action_client import ProfileActionClient, ActionClient
from subscription_manager.cli import system_exit
from subscription_manager.cli_command.cli import CliCommand, handle_exception
from subscription_manager.cli_command.list import show_autosubscribe_output
from subscription_manager.exceptions import ExceptionMapper
from subscription_manager.i18n import ugettext as _
from subscription_manager.packageprofilelib import PackageProfileActionInvoker
from subscription_manager.syspurposelib import save_sla_to_syspurpose_metadata
from subscription_manager.utils import is_simple_content_access, get_current_owner

log = logging.getLogger(__name__)


class AttachCommand(CliCommand):
    def __init__(self):
        super(AttachCommand, self).__init__(self._command_name(), self._short_description(), self._primary())

        self.product = None
        self.substoken = None
        self.auto_attach = True
        self.parser.add_argument(
            "--pool",
            dest="pool",
            action="append",
            help=_("The ID of the pool to attach (can be specified more than once)"),
        )
        self.parser.add_argument(
            "--quantity",
            dest="quantity",
            type=int,
            help=_("Number of subscriptions to attach. May not be used with an auto-attach."),
        )
        self.parser.add_argument(
            "--auto",
            action="store_true",
            help=_(
                "Automatically attach the best-matched compatible subscriptions to this system. "
                "This is the default action."
            ),
        )
        self.parser.add_argument(
            "--servicelevel",
            dest="service_level",
            help=_(
                "Automatically attach only subscriptions matching the specified service level; "
                "only used with --auto"
            ),
        )
        self.parser.add_argument(
            "--file",
            dest="file",
            help=_(
                "A file from which to read pool IDs. If a hyphen is provided, pool IDs will be "
                "read from stdin."
            ),
        )

        # re bz #864207
        _("All installed products are covered by valid entitlements.")
        _("No need to update subscriptions at this time.")

    def _read_pool_ids(self, f):
        if not self.options.pool:
            self.options.pool = []

        for line in fileinput.input(f):
            for pool in filter(bool, re.split(r"\s+", line.strip())):
                self.options.pool.append(pool)

    def _short_description(self):
        return _(
            "Attach a specified subscription to the registered system, when system does not use "
            "Simple Content Access mode"
        )

    def _command_name(self):
        return "attach"

    def _primary(self):
        return True

    def _validate_options(self):
        if self.options.pool or self.options.file:
            if self.options.auto:
                system_exit(os.EX_USAGE, _("Error: --auto may not be used when specifying pools."))
            if self.options.service_level:
                system_exit(
                    os.EX_USAGE, _("Error: The --servicelevel option cannot be used when specifying pools.")
                )

        # Quantity must be positive
        if self.options.quantity is not None:
            if self.options.quantity <= 0:
                system_exit(os.EX_USAGE, _("Error: Quantity must be a positive integer."))
            elif self.options.auto or not (self.options.pool or self.options.file):
                system_exit(os.EX_USAGE, _("Error: --quantity may not be used with an auto-attach"))

        # If a pools file was specified, process its contents and append it to options.pool
        if self.options.file:
            self.options.file = os.path.expanduser(self.options.file)
            if self.options.file == "-" or os.path.isfile(self.options.file):
                self._read_pool_ids(self.options.file)

                if len(self.options.pool) < 1:
                    if self.options.file == "-":
                        system_exit(os.EX_DATAERR, _("Error: Received data does not contain any pool IDs."))
                    else:
                        system_exit(
                            os.EX_DATAERR,
                            _('Error: The file "{file}" does not contain any pool IDs.').format(
                                file=self.options.file
                            ),
                        )
            else:
                system_exit(
                    os.EX_DATAERR,
                    _('Error: The file "{file}" does not exist or cannot be read.').format(
                        file=self.options.file
                    ),
                )

    def _print_ignore_attach_message(self):
        """
        Print message about ignoring attach request
        :return: None
        """
        owner = get_current_owner(self.cp, self.identity)
        owner_id = owner["key"]
        print(
            _(
                "Ignoring the request to attach. "
                'Attaching subscriptions is disabled for organization "{owner_id}" '
                "because Simple Content Access (SCA) is enabled."
            ).format(owner_id=owner_id)
        )

    def _do_command(self):
        """
        Executes the command.
        """
        self.assert_should_be_registered()
        self._validate_options()

        # --pool or --file turns off default auto attach
        if self.options.pool or self.options.file:
            self.auto_attach = False

        # Do not try to do auto-attach, when simple content access mode is used
        # BZ: https://bugzilla.redhat.com/show_bug.cgi?id=1826300
        #
        if is_simple_content_access(uep=self.cp, identity=self.identity):
            if self.auto_attach is True:
                self._print_ignore_auto_attach_message()
            else:
                self._print_ignore_attach_message()
            return 0

        installed_products_num = 0
        return_code = 0
        report = None

        # TODO: change to if self.auto_attach: else: pool/file stuff
        try:
            cert_action_client = ActionClient(skips=[PackageProfileActionInvoker])
            cert_action_client.update()
            cert_update = True

            attach_service = attach.AttachService(self.cp)
            if self.options.pool:
                subscribed = False

                for pool in self.options.pool:
                    # odd html strings will cause issues, reject them here.
                    if pool.find("#") >= 0:
                        system_exit(os.EX_USAGE, _("Please enter a valid numeric pool ID."))

                    try:
                        ents = attach_service.attach_pool(pool, self.options.quantity)
                        # Usually just one, but may as well be safe:
                        for ent in ents:
                            pool_json = ent["pool"]
                            print(
                                _("Successfully attached a subscription for: {name}").format(
                                    name=pool_json["productName"]
                                )
                            )
                            log.debug(
                                "Attached a subscription for {name}".format(name=pool_json["productName"])
                            )
                            subscribed = True
                    except connection.RestlibException as re:
                        log.exception(re)

                        exception_mapper = ExceptionMapper()
                        mapped_message = exception_mapper.get_message(re)

                        if re.code == 403:
                            print(mapped_message)  # already subscribed.
                        elif re.code == 400 or re.code == 404:
                            print(mapped_message)  # no such pool.
                        else:
                            system_exit(os.EX_SOFTWARE, mapped_message)  # some other error.. don't try again
                if not subscribed:
                    return_code = 1
            # must be auto
            else:
                installed_products_num = len(products.InstalledProducts(self.cp).list())
                # if we are green, we don't need to go to the server
                self.sorter = inj.require(inj.CERT_SORTER)

                if self.sorter.is_valid():
                    if not installed_products_num:
                        print(_("No Installed products on system. " "No need to attach subscriptions."))
                    else:
                        print(
                            _(
                                "All installed products are covered by valid entitlements. "
                                "No need to update subscriptions at this time."
                            )
                        )
                    cert_update = False
                else:
                    # If service level specified, make an additional request to
                    # verify service levels are supported on the server:
                    if self.options.service_level:
                        consumer = self.cp.getConsumer(self.identity.uuid)
                        if "serviceLevel" not in consumer:
                            system_exit(
                                os.EX_UNAVAILABLE,
                                _(
                                    "Error: The --servicelevel option is not "
                                    "supported by the server. Did not "
                                    "complete your request."
                                ),
                            )

                    attach_service.attach_auto(self.options.service_level)
                    if self.options.service_level is not None:
                        #  RHBZ 1632797 we should only save the sla if the sla was actually
                        #  specified. The uep and consumer_uuid are None, because service_level was sent
                        # to candlepin server using attach_service.attach_auto()
                        save_sla_to_syspurpose_metadata(
                            uep=None, consumer_uuid=None, service_level=self.options.service_level
                        )
                        print(_("Service level set to: {}").format(self.options.service_level))

            if cert_update:
                report = self.entcertlib.update()

            profile_action_client = ProfileActionClient()
            profile_action_client.update()

            if report and report.exceptions():
                print(_("Entitlement Certificate(s) update failed due to the following reasons:"))
                for e in report.exceptions():
                    print("\t-", str(e))
            elif self.auto_attach:
                if not installed_products_num:
                    return_code = 1
                else:
                    self.sorter.force_cert_check()
                    # Make sure that we get fresh status of installed products
                    status_cache = inj.require(inj.ENTITLEMENT_STATUS_CACHE)
                    status_cache.load_status(
                        self.sorter.cp_provider.get_consumer_auth_cp(),
                        self.sorter.identity.uuid,
                        self.sorter.on_date,
                    )
                    self.sorter.load()
                    # run this after entcertlib update, so we have the new entitlements
                    return_code = show_autosubscribe_output(self.cp, self.identity)

        except Exception as e:
            handle_exception("Unable to attach: {e}".format(e=e), e)

        # it is okay to call this no matter what happens above,
        # it's just a notification to perform a check
        self._request_validity_check()

        return return_code



from .fixture import SubManFixture, open_mock
from unittest import mock
from subscription_manager import syspurposelib
from subscription_manager.syspurposelib import SyspurposeSyncActionCommand, SyspurposeSyncActionReport
import json


class SyspurposeLibTests(SubManFixture):
    """
    Tests various functions of the syspurposelib module.
    """

    def setUp(self):
        super(SyspurposeLibTests, self).setUp()

        patch1 = mock.patch("subscription_manager.syspurposelib.json", wraps=json)
        self.mock_json = patch1.start()
        self.addCleanup(patch1.stop)

        patch2 = mock.patch.object(syspurposelib, "USER_SYSPURPOSE", new="/test/value")
        self.mock_sp_path = patch2.start()
        self.addCleanup(patch2.stop)

    def test_write_syspurpose(self):
        """
        Test that shows that the write_syspurpose method uses the SyspurposeStore when possible.
        :return:
        """
        test_values = {
            "role": "Test Role",
            "addons": ["Some addon"],
            "service_level_agreement": "Premium",
            "usage": "Dev",
        }

        # First mock out the SyspurposeStore
        with mock.patch.object(syspurposelib, "SyncedStore") as sp_store:
            result = syspurposelib.write_syspurpose(test_values)

            sp_store.assert_called_with(None)
            sp_store.return_value.update_local.assert_called_with(test_values)

            self.assertEqual(result, True)

    def test_write_syspurpose_with_no_syspurpose_store(self):
        """
        Test that shows the backup method of writing to the syspurpose.json file works.
        :return:
        """
        test_values = {
            "role": "Test Role",
            "addons": ["Some addon"],
            "service_level_agreement": "Premium",
            "usage": "Dev",
        }

        with mock.patch.object(syspurposelib, "SyncedStore", new=None):
            with open_mock() as mock_open:
                result = syspurposelib.write_syspurpose(test_values)
                self.mock_json.dump.assert_called_with(test_values, mock_open, ensure_ascii=True, indent=2)
                self.assertEqual(result, True)
                self.assertEqual(json.loads(mock_open.content_out()), test_values)

    def test_write_syspurpose_with_no_syspurpose_store_and_os_error(self):
        """
        Test that shows how the write_syspurpose method works with an error opening the file
        :return:
        """
        test_values = {
            "role": "Test Role",
            "addons": ["Some addon"],
            "service_level_agreement": "Premium",
            "usage": "Dev",
        }

        self.mock_json.dump.side_effect = OSError

        with mock.patch.object(syspurposelib, "SyncedStore", new=None):
            with open_mock() as mock_open:
                result = syspurposelib.write_syspurpose(test_values)
                self.mock_json.dump.assert_called_with(test_values, mock_open, ensure_ascii=True, indent=2)
                self.assertEqual(result, False)

                # We expect the file not to have been modified.
                self.assertEqual(mock_open.content_out(), "")


class SyspurposeSyncActionCommandTests(SubManFixture):
    def setUp(self):
        super(SyspurposeSyncActionCommandTests, self).setUp()

        self.command = SyspurposeSyncActionCommand()
        # Some mock data to use in tests
        self.remote_sp = {
            "role": "The best role",
            "addOns": ["Super shiny Addon 1"],
            "serviceLevel": "Topmost",
            "usage": "Deviousness",
        }
        self.base = {
            "role": self.remote_sp["role"],
            "addons": self.remote_sp["addOns"],
            "service_level_agreement": self.remote_sp["serviceLevel"],
            "usage": self.remote_sp["usage"],
        }
        self.local_sp = {
            "role": "Some Other Role",
            "addons": self.remote_sp["addOns"],
            "service_level_agreement": self.remote_sp["serviceLevel"],
            "usage": self.remote_sp["usage"],
        }

    def test_perform(self):
        """
        Super simple test to show that the perform method is running sync.
        :return:
        """
        with mock.patch("subscription_manager.syspurposelib.SyncedStore") as store:
            result = self.command.perform()
            store.return_value.sync.assert_called_once()
            self.assertTrue(isinstance(result, SyspurposeSyncActionReport))



import json
import sys
from contextlib import ExitStack

from ..test_managercli import TestCliCommand
from subscription_manager import syspurposelib
from subscription_manager import managercli

from ..fixture import Capture

from unittest.mock import patch


class TestAddonsCommand(TestCliCommand):
    command_class = managercli.AddonsCommand

    def _set_syspurpose(self, syspurpose):
        """
        Set the mocked out syspurpose to the given dictionary of values.
        Assumes it is called after syspurposelib.USER_SYSPURPOSE is mocked out.
        :param syspurpose: A dict of values to be set as the syspurpose
        :return: None
        """
        with open(syspurposelib.USER_SYSPURPOSE, "w") as sp_file:
            json.dump(syspurpose, sp_file, ensure_ascii=True)

    def setUp(self):
        syspurpose_patch = patch("syspurpose.files.SyncedStore")
        sp_patch = syspurpose_patch.start()
        self.addCleanup(sp_patch.stop)
        super(TestAddonsCommand, self).setUp()
        argv_patcher = patch.object(sys, "argv", ["subscription-manager", "addons"])
        argv_patcher.start()
        self.addCleanup(argv_patcher.stop)
        syspurposelib.USER_SYSPURPOSE = self.write_tempfile("{}").name

    def tearDown(self):
        super(TestAddonsCommand, self).tearDown()
        syspurposelib.USER_SYSPURPOSE = "/etc/rhsm/syspurpose/syspurpose.json"

    def test_view(self):
        self._test_no_exception([])

    def test_add(self):
        self._test_no_exception(["--add", "test"])

    def test_add_and_remove(self):
        self._test_exception(["--add", "test", "--remove", "something_else"])

    def test_remove(self):
        self._test_no_exception(["--remove", "test"])

    def test_unset(self):
        self._test_no_exception(["--unset"])

    def test_unset_and_add_and_remove(self):
        self._test_exception(["--add", "test", "--remove", "item", "--unset"])

    def test_add_valid_value(self):
        with patch.object(self.cc, "_get_valid_fields") as mock_get_valid_fields:
            mock_get_valid_fields.return_value = {"addons": ["ADDON1", "ADDON3", "ADDON2"]}
            self.assertTrue(self.cc._is_provided_value_valid("ADDON1"))
            with Capture() as cap:
                self.assertEqual(self.cc._are_provided_values_valid(["ADDON1"]), [])
            self.assertNotIn("Warning: Provided value", cap.out)

    def test_add_invalid_value(self):
        with patch.object(self.cc, "_get_valid_fields") as mock_get_valid_fields:
            mock_get_valid_fields.return_value = {"addons": ["ADDON1", "ADDON3", "ADDON2"]}
            self.assertFalse(self.cc._is_provided_value_valid("test"))
            with Capture() as cap:
                self.assertEqual(self.cc._are_provided_values_valid(["test"]), ["test"])
            self.assertIn("Warning: Provided value", cap.out)

    def test_no_valid_values(self):
        with ExitStack() as stack:
            mock_get_valid_fields = stack.enter_context(patch.object(self.cc, "_get_valid_fields"))
            cap = stack.enter_context(Capture())
            mock_get_valid_fields.return_value = {"addons": []}
            self.assertFalse(self.cc._is_provided_value_valid("test"))
            self.assertIn("Warning: This organization does not have", cap.out)



# Copyright (c) 2016 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
import os
import shutil

from distutils import log, dir_util
from distutils.command.build_py import build_py as _build_py
from distutils.spawn import spawn

from build_ext.utils import Utils, BaseCommand


# Courtesy http://wiki.maemo.org/Internationalize_a_Python_application
class BuildTrans(BaseCommand):
    description = "Compile .po files into .mo files"
    user_options = _build_py.user_options + [
        ("lint", "l", "check po files only"),
    ]

    boolean_options = ["lint"]
    app_name = "rhsm"

    def initialize_options(self):
        self.build_base = None
        self.lint = None

    def finalize_options(self):
        self.set_undefined_options("build", ("build_base", "build_base"))

    def compile(self, src, dest):
        log.debug("Compiling %s" % src)
        if self.lint:
            dest = "/dev/null"

        cmd = ["msgfmt", "--check", "--statistics", "-o", dest, src]
        spawn(cmd)

    def run(self):
        for po_file in Utils.find_files_of_type("po", "*.po"):
            lang = os.path.basename(po_file)[:-3]
            dest_path = os.path.join(self.build_base, "locale", lang, "LC_MESSAGES")
            dest = os.path.join(dest_path, self.app_name + ".mo")
            Utils.run_if_new(po_file, dest, self.compile)


class UpdateTrans(BaseCommand):
    description = "Update .po files with msgmerge"
    user_options = _build_py.user_options + [
        ("key-file=", "k", "file to update"),
    ]

    def initialize_options(self):
        self.key_file = os.path.join(os.curdir, "po", "keys.pot")

    def merge(self, po_file, key_file):
        log.debug("Updating %s" % po_file)
        cmd = ["msgmerge", "-N", "--backup=none", "-U", po_file, self.key_file]
        spawn(cmd)

    def run(self):
        for po_file in Utils.find_files_of_type("po", "*.po"):
            self.merge(po_file, self.key_file)


class UniqTrans(BaseCommand):
    description = "Unify duplicate translations with msguniq"

    def run(self):
        for po_file in Utils.find_files_of_type("po", "*.po"):
            cmd = ["msguniq", po_file, "-o", po_file]
            spawn(cmd)


class Gettext(BaseCommand):
    description = "Extract strings to po files."
    user_options = _build_py.user_options + [
        ("lint", "l", "check po files only"),
        ("key-file=", "k", "file to write to"),
    ]

    boolean_options = ["lint"]

    def initialize_options(self):
        self.build_base = None
        self.lint = None
        self.key_file = os.path.join(os.curdir, "po", "keys.pot")

    def finalize_options(self):
        self.src_dirs = self.distribution.package_dir
        print()
        if self.lint:
            self.key_file = "/dev/null"

    def find_py(self):
        files = []

        for src in self.src_dirs.values():
            print(src)
            files.extend(list(Utils.find_files_of_type(src, "*.py")))

        files.extend(list(Utils.find_files_of_type("bin", "*")))
        return files

    def find_c(self):
        files = []

        for src in self.src_dirs.values():
            files.extend(list(Utils.find_files_of_type(src, "*.c", "*.h")))

        return files

    def _write_sources(self, manifest_file, find_function):
        with open(manifest_file, "w") as f:
            f.write("\n".join(find_function()))
            f.write("\n")

    def run(self):
        manifest_prefix = os.path.join(os.curdir, "po", "POTFILES")

        # Begin with a fresh key file
        dir_util.mkpath("tmp")
        if self.lint:
            tmp_key_file = self.key_file
        else:
            tmp_key_file = os.path.join("tmp", os.path.basename(self.key_file))

        cmd = [
            "xgettext",
            "--from-code=utf-8",
            "--add-comments=TRANSLATORS:",
            "--sort-by-file",
            "-o",
            tmp_key_file,
            "--package-name=rhsm",
        ]

        # These tuples contain a template for the file name that will contain a list of
        # all source files of a given type to translate, a function that finds all the
        # source files of a given type, the source language to send to xgettext, and a
        # list of any additional arguments.
        #
        # ProTip: Use extensions that won't be confused with source files.
        trans_types = [
            # The C source files require that we inform xgettext to extract strings from the
            # _() and N_() functions.
            ("%s.c_files", self.find_c, "C", ["-k_", "-kN_"]),
            ("%s.py_files", self.find_py, "Python", []),
        ]

        for manifest_template, search_func, language, other_options in trans_types:
            manifest = manifest_template % manifest_prefix
            self._write_sources(manifest, search_func)

            specific_opts = ["-f", manifest, "--language", language]
            specific_opts.extend(other_options)
            if os.path.exists(tmp_key_file):
                specific_opts.append("--join-existing")
            log.debug("Running %s" % " ".join(cmd + specific_opts))
            spawn(cmd + specific_opts)

        if not self.lint:
            shutil.copy2(tmp_key_file, self.key_file)

        # Delete the directory holding the temporary keys.pot
        shutil.rmtree("tmp")


class GettextWithArgparse(Gettext):
    def find_py(self):
        # Can't use super since we're descended from a old-style class
        files = Gettext.find_py(self)

        # We need to grab some strings out of argparse for translation
        import argparse

        argparse_source = "%s.py" % os.path.splitext(argparse.__file__)[0]
        if not os.path.exists(argparse_source):
            raise RuntimeError("Could not find argparse.py at %s" % argparse_source)
        files.append(argparse_source)
        return files



#
# Subscription manager command line utility.
#
# Copyright (c) 2021 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#
import logging
import os

import rhsm.connection as connection

from subscription_manager.cli import system_exit
from subscription_manager.cli_command.cli import (
    handle_exception,
    ERR_NOT_REGISTERED_CODE,
    ERR_NOT_REGISTERED_MSG,
)
from subscription_manager.cli_command.org import OrgCommand
from subscription_manager.cli_command.list import ENVIRONMENT_LIST
from subscription_manager.i18n import ugettext as _
from subscription_manager.i18n import ungettext
from subscription_manager.printing_utils import columnize, echo_columnize_callback
from subscription_manager.utils import get_supported_resources

from subscription_manager.injection import require, IDENTITY


log = logging.getLogger(__name__)
MULTI_ENV = "multi_environment"


class EnvironmentsCommand(OrgCommand):
    def __init__(self):
        shortdesc = _("Display the environments available for a user")
        self._org_help_text = _("specify organization for environment list, using organization key")

        super(EnvironmentsCommand, self).__init__("environments", shortdesc, False)
        self._add_url_options()
        self.parser.add_argument(
            "--set",
            dest="set",
            help=_("set an ordered comma-separated list of environments for this consumer"),
        )
        self.parser.add_argument(
            "--list",
            action="store_true",
            default=False,
            help=_("list all environments for the organization"),
        )
        self.parser.add_argument(
            "--list-enabled",
            action="store_true",
            dest="enabled",
            default=False,
            help=_("list the environments enabled for this consumer in order"),
        )
        self.parser.add_argument(
            "--list-disabled",
            action="store_true",
            dest="disabled",
            default=False,
            help=_("list the environments not enabled for this consumer"),
        )

    def _get_environments(self, org):
        return self.cp.getEnvironmentList(org)

    def _validate_options(self):
        if self.identity.is_valid():
            if self.options.org:
                system_exit(os.EX_USAGE, _("You may not specify an --org for environments when registered."))
        else:
            if self.options.enabled or self.options.disabled:
                system_exit(ERR_NOT_REGISTERED_CODE, ERR_NOT_REGISTERED_MSG)

    def _do_command(self):
        self._validate_options()
        supported_resources = get_supported_resources()
        if "environments" not in supported_resources:
            system_exit(os.EX_UNAVAILABLE, _("Error: Server does not support environments."))
        try:
            if self.options.token:
                self.cp = self.cp_provider.get_keycloak_auth_cp(self.options.token)
            else:
                if not self.options.enabled:
                    if self.options.username is None or self.options.password is None:
                        print(_("This operation requires user credentials"))
                    self.cp_provider.set_user_pass(self.username, self.password)
                    self.cp = self.cp_provider.get_basic_auth_cp()
            self.identity = require(IDENTITY)
            if self.options.set:
                self._set_environments()
            else:
                self._list_environments()
        except connection.RestlibException as re:
            log.exception(re)
            log.error("Error: Unable to retrieve environment list from server: {re}".format(re=re))

            system_exit(os.EX_SOFTWARE, re)
        except Exception as e:
            handle_exception(_("Error: Unable to retrieve environment list from server"), e)

    def _set_environments(self):
        """
        Updates the environments for the consumer if that is a capability at the server
        """
        if self.cp.has_capability(MULTI_ENV):
            if not self.identity.is_valid():
                system_exit(ERR_NOT_REGISTERED_CODE, ERR_NOT_REGISTERED_MSG)
            self.cp.updateConsumer(
                self.identity.uuid,
                environments=self._process_environments(
                    self.cp, self.cp.getOwner(self.identity.uuid)["key"], self.options
                ),
            )
            print(_("Environments updated."))
        else:
            system_exit(os.EX_UNAVAILABLE, _("Error: Server does not support environment updates."))

    def _list_environments(self):
        """
        List the environments based on the option selected in the command line
        enabled/disabled/all
        """
        if not self.cp.has_capability(MULTI_ENV) and (self.options.enabled or self.options.disabled):
            system_exit(os.EX_UNAVAILABLE, _("Error: Server does not support multi-environment operations."))
        environments = []
        if self.options.enabled:
            environments = self.cp.getConsumer(self.identity.uuid)["environments"] or []
        else:
            org_environments = self._get_environments(self.org)
            if self.options.disabled:
                consumer_id_list = []
                for env in self.cp.getConsumer(self.identity.uuid)["environments"]:
                    consumer_id_list.append(env["id"])
                for env in org_environments:
                    if env["id"] not in consumer_id_list:
                        environments.append(env)
            else:
                environments = org_environments

        if len(environments):
            print("+-------------------------------------------+")
            print("          {env}".format(env=_("Environments")))
            print("+-------------------------------------------+")
            for env in environments:
                print(
                    columnize(
                        ENVIRONMENT_LIST,
                        echo_columnize_callback,
                        env["name"],
                        env["description"] or "",
                    )
                    + "\n"
                )
        else:
            print(_("This list operation does not have any environments to report."))

    def _process_environments(self, admin_cp, owner_key, options):
        """
        Gather environment list from server and pass to method to
        validate the environment input against it
        """
        all_env_list = admin_cp.getEnvironmentList(owner_key)
        return check_set_environment_names(all_env_list, options.set)

    @property
    def org(self):
        """
        An override is needed here because we need to use the org from the
        identity for this command if this system is already registered
        """
        self.identity = require(IDENTITY)
        if self.identity.is_valid():
            self._org = self.cp.getOwner(self.identity.uuid)["key"]
            return self._org
        elif self.options.org:
            self._org = self.options.org
            return self._org
        else:
            if not self.options.username:
                self.options.username = self._username
            return super().org


def check_set_environment_names(all_env_list, name_string):
    """
    Checks the environment name(s) input for duplicates and
    inclusion in the full list
    """
    names = [name.strip() for name in name_string.split(",")]
    if len(names) > len(set(names)):
        system_exit(os.EX_DATAERR, _("Error: The same environment may not be listed more than once. "))

    all_names_ids = dict((environment["name"], environment["id"]) for environment in all_env_list)
    missing_names = [name for name in names if name not in all_names_ids.keys()]
    if len(missing_names) > 0:
        msg = ungettext(
            "No such environment: {names}", "No such environments: {names}", len(missing_names)
        ).format(names=", ".join(missing_names))
        system_exit(os.EX_DATAERR, msg)

    return ",".join([all_names_ids[name] for name in names])



# Copyright (c) 2017 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
import datetime
from unittest import mock

from test.rhsmlib.base import InjectionMockingTest

from subscription_manager import injection as inj
from subscription_manager.identity import Identity
from subscription_manager.cert_sorter import CertSorter
from subscription_manager.reasons import Reasons
from subscription_manager.certdirectory import EntitlementDirectory
from subscription_manager.cp_provider import CPProvider
from subscription_manager.cache import AvailableEntitlementsCache

from rhsm import connection

from rhsmlib.services import exceptions
from rhsmlib.services.entitlement import EntitlementService


class TestEntitlementService(InjectionMockingTest):
    def setUp(self):
        super(TestEntitlementService, self).setUp()
        self.mock_identity = mock.Mock(spec=Identity, name="Identity").return_value
        self.mock_cp = mock.Mock(spec=connection.UEPConnection, name="UEPConnection").return_value
        self.mock_sorter_class = mock.Mock(spec=CertSorter, name="CertSorter")
        self.mock_ent_dir = mock.Mock(spec=EntitlementDirectory, name="EntitlementDirectory").return_value
        self.mock_cache_avail_ent = mock.Mock(
            spec=AvailableEntitlementsCache, name="AvailableEntitlements"
        ).return_value
        self.mock_cache_avail_ent.get_not_obsolete_data = mock.Mock(return_value=[])
        self.mock_cache_avail_ent.timeout = mock.Mock(return_value=10.0)
        self.mock_provider = mock.Mock(spec=CPProvider, name="CPProvider")
        self.mock_provider.get_consumer_auth_cp.return_value = mock.Mock(name="MockCP")

    def injection_definitions(self, *args, **kwargs):
        if args[0] == inj.IDENTITY:
            return self.mock_identity
        elif args[0] == inj.CERT_SORTER:
            # This sleight of hand is needed because we want to check the arguments given
            # when the cert sorter is instantiated.
            instance = self.mock_sorter_class(*args[1:])
            self.mock_sorter_class.return_value = instance
            return instance
        elif args[0] == inj.ENT_DIR:
            return self.mock_ent_dir
        elif args[0] == inj.AVAILABLE_ENTITLEMENT_CACHE:
            return self.mock_cache_avail_ent
        elif args[0] == inj.CP_PROVIDER:
            return self.mock_provider
        else:
            return None

    def test_get_status(self):
        # Prime the injected dependencies built in injection_definitions
        service = EntitlementService()

        self.mock_identity.is_valid.return_value = True

        mock_reasons = mock.Mock(spec=Reasons, name="Reasons").return_value
        mock_reasons.get_name_message_map.return_value = {"RHEL": ["Not supported by a valid subscription"]}
        mock_reasons.get_reason_ids_map.return_value = {"69": {"key": "NOTCOVERED", "product_name": "RHEL"}}

        mock_sorter = self.mock_sorter_class.return_value
        mock_sorter.get_system_status.return_value = "Invalid"
        mock_sorter.get_system_status_id.return_value = "invalid"
        mock_sorter.reasons = mock_reasons
        mock_sorter.is_valid.return_value = False

        expected_value = {
            "status": "Invalid",
            "status_id": "invalid",
            "reasons": {"RHEL": ["Not supported by a valid subscription"]},
            "reason_ids": {"69": {"key": "NOTCOVERED", "product_name": "RHEL"}},
            "valid": False,
        }

        self.assertEqual(expected_value, service.get_status())

    def test_get_status_on_date(self):
        # Prime the injected dependencies with those built in injection_definitions
        service = EntitlementService()

        # Verify the cert_sorter was constructed for a specific date
        mock_sorter = self.mock_sorter_class.return_value

        self.mock_identity.is_valid.return_value = True

        mock_reasons = mock.Mock(spec=Reasons, name="Reasons").return_value
        mock_reasons.get_name_message_map.return_value = {"RHEL": ["Not supported by a valid subscription"]}
        mock_reasons.get_reason_ids_map.return_value = {"69": {"key": "NOTCOVERED", "product_name": "RHEL"}}

        mock_sorter.get_system_status.return_value = "Invalid"
        mock_sorter.get_system_status_id.return_value = "invalid"
        mock_sorter.reasons = mock_reasons
        mock_sorter.is_valid.return_value = False

        expected_value = {
            "status": "Invalid",
            "status_id": "invalid",
            "reasons": {"RHEL": ["Not supported by a valid subscription"]},
            "reason_ids": {"69": {"key": "NOTCOVERED", "product_name": "RHEL"}},
            "valid": False,
        }

        self.assertEqual(expected_value, service.get_status("some_date"))
        self.mock_sorter_class.assert_called_once_with("some_date")

    def _build_options(
        self,
        pool_subsets=None,
        matches=None,
        pool_only=None,
        match_installed=None,
        no_overlap=None,
        service_level=None,
        show_all=None,
        on_date=None,
    ):
        return {
            "pool_subsets": pool_subsets,
            "matches": matches,
            "pool_only": pool_only,
            "match_installed": match_installed,
            "no_overlap": no_overlap,
            "service_level": service_level,
            "show_all": show_all,
            "on_date": on_date,
        }

    def test_get_status_when_unregistered(self):
        # Prime the injected dependencies with those built in injection_definitions
        service = EntitlementService()

        self.mock_identity.is_valid.return_value = False
        expected_value = {
            "status": "Unknown",
            "status_id": "unknown",
            "reasons": {},
            "reason_ids": {},
            "valid": False,
        }
        self.assertEqual(expected_value, service.get_status())

    def test_only_accepts_correct_pool_subsets(self):
        service = EntitlementService()
        options = self._build_options(pool_subsets=["foo"])
        with self.assertRaisesRegex(exceptions.ValidationError, r".*invalid listing type.*"):
            service.validate_options(options)

    def test_show_all_requires_available(self):
        service = EntitlementService()
        options = self._build_options(pool_subsets=["installed"], show_all=True)
        with self.assertRaisesRegex(exceptions.ValidationError, r".*only applicable with --available"):
            service.validate_options(options)
        options["pool_subsets"].append("available")
        service.validate_options(options)

    def test_on_date_requires_available(self):
        service = EntitlementService()
        on_date = datetime.date.today().strftime("%Y-%m-%d")
        options = self._build_options(pool_subsets=["installed", "consumed"], on_date=on_date)
        with self.assertRaisesRegex(exceptions.ValidationError, r".*only applicable with --available"):
            service.validate_options(options)
        options["pool_subsets"].append("available")
        service.validate_options(options)

    def test_service_level_requires_consumed_or_available(self):
        service = EntitlementService()
        options = self._build_options(pool_subsets=["installed"], service_level="foo")
        with self.assertRaisesRegex(exceptions.ValidationError, r".*only applicable with --available"):
            service.validate_options(options)
        options["pool_subsets"].append("available")
        service.validate_options(options)

    def test_match_installed_requires_available(self):
        service = EntitlementService()
        options = self._build_options(pool_subsets=["installed", "consumed"], match_installed=True)
        with self.assertRaisesRegex(exceptions.ValidationError, r".*only applicable with --available"):
            service.validate_options(options)
        options["pool_subsets"].append("available")
        service.validate_options(options)

    def test_no_overlap_requires_available(self):
        service = EntitlementService()
        options = self._build_options(pool_subsets=["installed", "consumed"], no_overlap=True)
        with self.assertRaisesRegex(exceptions.ValidationError, r".*only applicable with --available"):
            service.validate_options(options)
        options["pool_subsets"].append("available")
        service.validate_options(options)

    def test_pool_only_requires_consumed_or_available(self):
        service = EntitlementService()
        options = self._build_options(pool_subsets=["installed"], pool_only=True)
        with self.assertRaisesRegex(exceptions.ValidationError, r".*only applicable with --available"):
            service.validate_options(options)
        options["pool_subsets"].append("available")
        service.validate_options(options)

    def test_available_requires_registration(self):
        service = EntitlementService()
        self.mock_identity.is_valid.return_value = False
        options = self._build_options(pool_subsets=["available"])
        with self.assertRaisesRegex(exceptions.ValidationError, r".*not registered.*"):
            service.validate_options(options)

    @mock.patch("rhsmlib.services.entitlement.managerlib")
    def test_filter_only_specified_service_level(self, mock_managerlib):
        service = EntitlementService()
        pools = [{"service_level": "Level1"}, {"service_level": "Level2"}, {"service_level": "Level3"}]
        mock_managerlib.get_available_entitlements.return_value = pools

        filtered = service.get_available_pools(service_level="Level2")

        self.assertEqual(1, len(filtered))
        self.assertEqual("Level2", filtered[0]["service_level"])

    @mock.patch("rhsmlib.services.entitlement.managerlib")
    def test_pagged_result(self, mock_managerlib):
        service = EntitlementService()
        pools = [
            {"id": "ff8080816ea20fb9016ea21283ab02e0"},
            {"id": "ff8080816ea20fb9016ea21283ab02e1"},
            {"id": "ff8080816ea20fb9016ea21283ab02e2"},
            {"id": "ff8080816ea20fb9016ea21283ab02e3"},
            {"id": "ff8080816ea20fb9016ea21283ab02e4"},
        ]
        mock_managerlib.get_available_entitlements.return_value = pools

        filtered = service.get_available_pools(page=1, items_per_page=3)

        self.assertEqual(2, len(filtered))
        self.assertEqual("ff8080816ea20fb9016ea21283ab02e3", filtered[0]["id"])
        self.assertEqual(1, filtered[0]["page"])
        self.assertEqual(3, filtered[0]["items_per_page"])

    @mock.patch("rhsmlib.services.entitlement.managerlib")
    def test_no_pagged_result(self, mock_managerlib):
        service = EntitlementService()
        pools = [
            {"id": "ff8080816ea20fb9016ea21283ab02e0"},
            {"id": "ff8080816ea20fb9016ea21283ab02e1"},
            {"id": "ff8080816ea20fb9016ea21283ab02e2"},
            {"id": "ff8080816ea20fb9016ea21283ab02e3"},
            {"id": "ff8080816ea20fb9016ea21283ab02e4"},
        ]
        mock_managerlib.get_available_entitlements.return_value = pools

        filtered = service.get_available_pools(page=0, items_per_page=0)

        self.assertEqual(5, len(filtered))
        self.assertEqual("ff8080816ea20fb9016ea21283ab02e0", filtered[0]["id"])
        self.assertNotIn("page", filtered[0])
        self.assertNotIn("items_per_page", filtered[0])

    @mock.patch("rhsmlib.services.entitlement.managerlib")
    def test_pagged_result_too_big_page_value(self, mock_managerlib):
        service = EntitlementService()
        pools = [
            {"id": "ff8080816ea20fb9016ea21283ab02e0"},
            {"id": "ff8080816ea20fb9016ea21283ab02e1"},
            {"id": "ff8080816ea20fb9016ea21283ab02e2"},
            {"id": "ff8080816ea20fb9016ea21283ab02e3"},
            {"id": "ff8080816ea20fb9016ea21283ab02e4"},
        ]
        mock_managerlib.get_available_entitlements.return_value = pools

        filtered = service.get_available_pools(page=10, items_per_page=3)

        self.assertEqual(0, len(filtered))

    @mock.patch("rhsmlib.services.entitlement.managerlib")
    def test_no_pool_with_specified_filter(self, mock_managerlib):
        service = EntitlementService()
        pools = [{"service_level": "Level1"}]
        mock_managerlib.get_available_entitlements.return_value = pools

        filtered = service.get_available_pools(service_level="NotFound")
        self.assertEqual(0, len(filtered))

    def test_remove_all_pools(self):
        """
        Test of removing all pools
        """
        ent_service = EntitlementService(self.mock_cp)
        ent_service.entcertlib = mock.Mock().return_value
        ent_service.entcertlib.update = mock.Mock()
        ent_service.cp.unbindAll = mock.Mock(return_value="[]")

        response = ent_service.remove_all_entitlements()
        self.assertEqual(response, "[]")

    def test_remove_all_pools_by_id(self):
        """
        Test of removing all pools by IDs of pool
        """
        ent_service = EntitlementService(self.mock_cp)
        ent_service.cp.unbindByPoolId = mock.Mock()
        ent_service.entitlement_dir.list_serials_for_pool_ids = mock.Mock(
            return_value={
                "4028fa7a5dea087d015dea0b025003f6": ["6219625278114868779"],
                "4028fa7a5dea087d015dea0adf560152": ["3573249574655121394"],
            }
        )
        ent_service.entcertlib = mock.Mock().return_value
        ent_service.entcertlib.update = mock.Mock()

        removed_pools, unremoved_pools, removed_serials = ent_service.remove_entitlements_by_pool_ids(
            ["4028fa7a5dea087d015dea0b025003f6", "4028fa7a5dea087d015dea0adf560152"]
        )

        expected_removed_serials = ["6219625278114868779", "3573249574655121394"]
        expected_removed_pools = ["4028fa7a5dea087d015dea0b025003f6", "4028fa7a5dea087d015dea0adf560152"]

        self.assertEqual(expected_removed_serials, removed_serials)
        self.assertEqual(expected_removed_pools, removed_pools)
        self.assertEqual([], unremoved_pools)

    def test_remove_dupli_pools_by_id(self):
        """
        Test of removing pools specified with duplicities
        (one pool id is set twice)
        """
        ent_service = EntitlementService(self.mock_cp)
        ent_service.cp.unbindByPoolId = mock.Mock()
        ent_service.entitlement_dir.list_serials_for_pool_ids = mock.Mock(
            return_value={
                "4028fa7a5dea087d015dea0b025003f6": ["6219625278114868779"],
                "4028fa7a5dea087d015dea0adf560152": ["3573249574655121394"],
            }
        )
        ent_service.entcertlib = mock.Mock().return_value
        ent_service.entcertlib.update = mock.Mock()

        removed_pools, unremoved_pools, removed_serials = ent_service.remove_entitlements_by_pool_ids(
            [
                "4028fa7a5dea087d015dea0b025003f6",
                "4028fa7a5dea087d015dea0b025003f6",
                "4028fa7a5dea087d015dea0adf560152",
            ]
        )

        expected_removed_serials = ["6219625278114868779", "3573249574655121394"]
        expected_removed_pools = ["4028fa7a5dea087d015dea0b025003f6", "4028fa7a5dea087d015dea0adf560152"]

        self.assertEqual(expected_removed_serials, removed_serials)
        self.assertEqual(expected_removed_pools, removed_pools)
        self.assertEqual([], unremoved_pools)

    def test_remove_some_pools_by_id(self):
        """
        Test of removing only some pools, because one pool ID is not valid
        """
        ent_service = EntitlementService(self.mock_cp)

        def stub_unbind(uuid, pool_id):
            if pool_id == "does_not_exist_d015dea0adf560152":
                raise connection.RestlibException(400, "Error")

        ent_service.cp.unbindByPoolId = mock.Mock(side_effect=stub_unbind)
        ent_service.entitlement_dir.list_serials_for_pool_ids = mock.Mock(
            return_value={
                "4028fa7a5dea087d015dea0b025003f6": ["6219625278114868779"],
                "4028fa7a5dea087d015dea0adf560152": ["3573249574655121394"],
            }
        )
        ent_service.entcertlib = mock.Mock().return_value
        ent_service.entcertlib.update = mock.Mock()

        removed_pools, unremoved_pools, removed_serials = ent_service.remove_entitlements_by_pool_ids(
            ["4028fa7a5dea087d015dea0b025003f6", "does_not_exist_d015dea0adf560152"]
        )

        expected_removed_serials = ["6219625278114868779"]
        expected_removed_pools = ["4028fa7a5dea087d015dea0b025003f6"]
        expected_unremoved_pools = ["does_not_exist_d015dea0adf560152"]

        self.assertEqual(expected_removed_serials, removed_serials)
        self.assertEqual(expected_removed_pools, removed_pools)
        self.assertEqual(expected_unremoved_pools, unremoved_pools)

    def test_remove_all_pools_by_serial(self):
        """
        Test of removing all pools by serial numbers
        """
        ent_service = EntitlementService(self.mock_cp)
        ent_service.cp.unbindBySerial = mock.Mock()

        ent_service.entcertlib = mock.Mock().return_value
        ent_service.entcertlib.update = mock.Mock()

        removed_serial, unremoved_serials = ent_service.remove_entitlements_by_serials(
            ["6219625278114868779", "3573249574655121394"]
        )

        expected_removed_serials = ["6219625278114868779", "3573249574655121394"]

        self.assertEqual(expected_removed_serials, removed_serial)
        self.assertEqual([], unremoved_serials)

    def test_remove_dupli_pools_by_serial(self):
        """
        Test of removing pools specified with duplicities
        (one serial number is set twice)
        """
        ent_service = EntitlementService(self.mock_cp)
        ent_service.cp.unbindBySerial = mock.Mock()

        ent_service.entcertlib = mock.Mock().return_value
        ent_service.entcertlib.update = mock.Mock()

        removed_serial, unremoved_serials = ent_service.remove_entitlements_by_serials(
            ["6219625278114868779", "6219625278114868779", "3573249574655121394"]
        )

        expected_removed_serials = ["6219625278114868779", "3573249574655121394"]

        self.assertEqual(expected_removed_serials, removed_serial)
        self.assertEqual([], unremoved_serials)

    def test_remove_some_pools_by_serial(self):
        """
        Test of removing some of pools by serial numbers, because one serial
        number is not valid.
        """
        ent_service = EntitlementService(self.mock_cp)

        def stub_unbind(uuid, serial):
            if serial == "does_not_exist_1394":
                raise connection.RestlibException(400, "Error")

        ent_service.cp.unbindBySerial = mock.Mock(side_effect=stub_unbind)

        ent_service.entcertlib = mock.Mock().return_value
        ent_service.entcertlib.update = mock.Mock()

        removed_serial, unremoved_serials = ent_service.remove_entitlements_by_serials(
            ["6219625278114868779", "does_not_exist_1394"]
        )

        expected_removed_serials = ["6219625278114868779"]
        expected_unremoved_serials = ["does_not_exist_1394"]

        self.assertEqual(expected_removed_serials, removed_serial)
        self.assertEqual(expected_unremoved_serials, unremoved_serials)

    def test_parse_valid_date(self):
        """
        Test parsing valid date
        """
        on_date = datetime.date.today().strftime("%Y-%m-%d")
        ent_service = EntitlementService(self.mock_cp)
        expected_result = datetime.datetime.strptime(on_date, "%Y-%m-%d")
        parsed_date = ent_service.parse_date(on_date)
        self.assertEqual(expected_result, parsed_date)

    def test_parse_invalid_date(self):
        """
        Test parsing invalid date (invalid format)
        """
        on_date = "2000-20-20"
        ent_service = EntitlementService(self.mock_cp)
        self.assertRaises(ValueError, ent_service.parse_date, on_date)

    def test_parse_yesterday(self):
        """
        Test parsing invalid date (past dates are not allowed)
        """
        yesterday = datetime.date.today() - datetime.timedelta(1)
        on_date = yesterday.strftime("%Y-%m-%d")
        ent_service = EntitlementService(self.mock_cp)
        self.assertRaises(ValueError, ent_service.parse_date, on_date)



import os
import tempfile

from ..test_managercli import TestCliProxyCommand
from subscription_manager import syspurposelib
from subscription_manager import managercli

from ..stubs import StubConsumerIdentity, StubUEP
from ..fixture import set_up_mock_sp_store

from unittest.mock import patch, Mock, call


class TestServiceLevelCommand(TestCliProxyCommand):
    command_class = managercli.ServiceLevelCommand

    def setUp(self):
        syspurpose_patch = patch("syspurpose.files.SyncedStore")
        sp_patch = syspurpose_patch.start()
        self.addCleanup(sp_patch.stop)
        TestCliProxyCommand.setUp(self)
        self.cc.consumerIdentity = StubConsumerIdentity
        self.cc.cp = StubUEP()
        # Set up syspurpose mocking, do not test functionality of other source tree.
        from subscription_manager import syspurposelib

        self.syspurposelib = syspurposelib
        self.syspurposelib.USER_SYSPURPOSE = self.write_tempfile("{}").name

        syspurpose_patch = patch("subscription_manager.cli_command.abstract_syspurpose.SyncedStore")
        self.mock_sp_store = syspurpose_patch.start()
        self.mock_sp_store, self.mock_sp_store_contents = set_up_mock_sp_store(self.mock_sp_store)
        self.addCleanup(syspurpose_patch.stop)

    def tearDown(self):
        super(TestServiceLevelCommand, self).tearDown()
        syspurposelib.USER_SYSPURPOSE = "/etc/rhsm/syspurpose/syspurpose.json"

    def test_main_server_url(self):
        server_url = "https://subscription.rhsm.redhat.com/subscription"
        self.cc.main(["--serverurl", server_url])

    def test_insecure(self):
        self.cc.main(["--insecure"])

    def test_show(self):
        self.cc.main(["--show"])

    def test_no_arg(self):
        self.cc.main([])

    def test_org_requires_list_error(self):
        try:
            self.cc.main(["--org", "one"])
            self.cc._validate_options()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_USAGE)

    def test_set_allows_list_good(self):
        self.cc.is_registered = Mock(return_value=False)
        self.cc.main(["--set", "two", "--org", "test"])
        self.cc._validate_options()

    def test_org_requires_list_good(self):
        self.cc.main(["--org", "one", "--list"])

    def test_list_with_one_org_no_prompt(self):
        owner_list = self.cc.cp.getOwnerList
        self.cc.cp.getOwnerList = Mock(return_value="test_org")
        self.cc.main(["--list"])
        self.cc.cp.getOwnerList = owner_list

    def test_service_level_supported(self):
        self.cc.cp.setConsumer({"serviceLevel": "Jarjar"})
        self.cc._set("JRJAR")

    @patch("subscription_manager.cli_command.service_level.SyncedStore")
    def test_service_level_set_creates_syspurpose_dir_and_file(self, mock_syspurpose):
        # create a mock /etc/rhsm/ directory, and set the value of a mock USER_SYSPURPOSE under that
        old_capabilities = self.cc.cp._capabilities
        self.cc.cp._capabilities = ["syspurpose"]
        self.cc.store = self.mock_sp_store()
        self.cc.store.get_cached_contents = Mock(return_value={})

        mock_syspurpose.read = Mock()
        mock_syspurpose.read.return_value = Mock()
        mock_syspurpose.return_value = self.mock_sp_store()

        mock_etc_rhsm_dir = tempfile.TemporaryDirectory()
        mock_syspurpose_file = os.path.join(mock_etc_rhsm_dir.name, "syspurpose/syspurpose.json")
        syspurposelib.USER_SYSPURPOSE = mock_syspurpose_file

        self.cc.store = self.mock_sp_store()
        self.cc.options = Mock()
        self.cc.options.set = "JRJAR"
        self.cc.set()

        self.cc.store.set.assert_has_calls([call("service_level_agreement", "JRJAR")])

        # make sure the sla has been persisted in syspurpose.json:
        contents = self.cc.store.get_local_contents()
        self.assertEqual(contents.get("service_level_agreement"), "JRJAR")
        self.cc.cp._capabilities = old_capabilities

    def test_old_service_level(self):
        self.cc.options = Mock()
        self.cc.cp.getConsumer = Mock(return_value={"serviceLevel": "foo"})
        self.cc.options.set = "JRJAR"
        self.cc.cp.updateConsumer = Mock()
        # 'syspurpose' is not in capabilities of server
        self.cc.set()
        self.cc.cp.updateConsumer.assert_has_calls(
            [call("fixture_identity_mock_uuid", service_level="JRJAR")]
        )

    def test_username_on_registered_system(self):
        """Argument --username cannot be used on registered system."""
        self.cc.is_registered = Mock(return_value=True)
        self.cc.options = Mock()
        self.cc.options.set = None
        self.cc.options.unset = None
        self.cc.options.to_add = None
        self.cc.options.to_remove = None
        self.cc.options.show = None
        self.cc.options.list = True
        self.cc.options.username = "admin"
        try:
            self.cc._validate_options()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_USAGE)

    def test_password_on_registered_system(self):
        """Argument --password cannot be used on registered system."""
        self.cc.is_registered = Mock(return_value=True)
        self.cc.options = Mock()
        self.cc.options.set = None
        self.cc.options.unset = None
        self.cc.options.to_add = None
        self.cc.options.to_remove = None
        self.cc.options.show = None
        self.cc.options.list = True
        self.cc.options.password = "secret"
        try:
            self.cc._validate_options()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_USAGE)

    def test_token_on_registered_system(self):
        """Argument --token cannot be used on registered system."""
        self.cc.is_registered = Mock(return_value=True)
        self.cc.options = Mock()
        self.cc.options.set = None
        self.cc.options.unset = None
        self.cc.options.to_add = None
        self.cc.options.to_remove = None
        self.cc.options.show = None
        self.cc.options.list = True
        self.cc.options.token = "TOKEN"
        try:
            self.cc._validate_options()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_USAGE)

    def test_org_on_registered_system(self):
        """Argument --org cannot be used on registered system."""
        self.cc.is_registered = Mock(return_value=True)
        self.cc.options = Mock()
        self.cc.options.set = None
        self.cc.options.unset = None
        self.cc.options.to_add = None
        self.cc.options.to_remove = None
        self.cc.options.show = None
        self.cc.options.list = True
        self.cc.options.org = "organization"
        try:
            self.cc._validate_options()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_USAGE)

    def test_serverurl_on_registered_system(self):
        """Argument --serverurl cannot be used on registered system."""
        self.cc.is_registered = Mock(return_value=True)
        self.cc.options = Mock()
        self.cc.options.set = None
        self.cc.options.unset = None
        self.cc.options.to_add = None
        self.cc.options.to_remove = None
        self.cc.options.show = None
        self.cc.options.list = True
        self.cc.options.org = None
        self.cc.options.server_url = "https://example.com"
        try:
            self.cc._validate_options()
        except SystemExit as e:
            self.assertEqual(e.code, os.EX_USAGE)



#
# Subscription manager command line utility.
#
# Copyright (c) 2021 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#
import getpass
import readline
import os

from subscription_manager.cli_command.cli import CliCommand
from subscription_manager.cli import system_exit
from subscription_manager.i18n import ugettext as _
from subscription_manager.utils import is_interactive


class UserPassCommand(CliCommand):
    """
    Abstract class for commands that require a username and password
    """

    def __init__(self, name, shortdesc=None, primary=False):
        super(UserPassCommand, self).__init__(name, shortdesc, primary)
        self._username = None
        self._password = None

        self.parser.add_argument(
            "--username",
            dest="username",
            help=_("username to use when authorizing against the server"),
        )
        self.parser.add_argument(
            "--password",
            dest="password",
            help=_("password to use when authorizing against the server"),
        )
        self.parser.add_argument(
            "--token",
            dest="token",
            help=_("token to use when authorizing against the server"),
        )

    @staticmethod
    def _get_username_and_password(username, password):
        """
        Safely get a username and password from the tty, without echoing.
        if either username or password are provided as arguments, they will
        not be prompted for. In a non-interactive session, the system exits with an error.
        """
        if not is_interactive():
            if not username:
                system_exit(
                    os.EX_USAGE,
                    _("Error: --username is a required parameter in non-interactive mode."),
                )
            if not password:
                system_exit(
                    os.EX_USAGE,
                    _("Error: --password is a required parameter in non-interactive mode."),
                )

        while not username:
            username = input(_("Username: "))
            readline.clear_history()
        while not password:
            password = getpass.getpass(_("Password: "))
        return username.strip(), password.strip()

    # lazy load the username and password, prompting for them if they weren't
    # given as options. this lets us not prompt if another option fails,
    # or we don't need them.
    @property
    def username(self):
        if not self._username:
            if self.options.token:
                self._username = self.cp_provider.token_username
                return self._username
            (self._username, self._password) = self._get_username_and_password(
                self.options.username, self.options.password
            )
        return self._username

    @property
    def password(self):
        if not self._password and not self.options.token:
            (self._username, self._password) = self._get_username_and_password(
                self.options.username, self.options.password
            )
        return self._password



# Copyright (c) 2018 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

"""
Utility methods for the syspurpose command
"""

import io
import json
import os
import errno
import logging

log = logging.getLogger(__name__)


def create_dir(path: str) -> bool:
    """
    Attempts to create the path given (less any file)
    :param path: path
    :return: True if changes were made, false otherwise
    """
    try:
        os.makedirs(path, mode=0o755)
    except OSError as e:
        if e.errno == errno.EEXIST:
            # If the directory exists no changes necessary
            return False
        if e.errno == errno.EACCES:
            log.error("Cannot create directory {}".format(path))
    return True


def create_file(path: str, contents: dict) -> bool:
    """
    Attempts to create a file, with the given contents
    :param path: The desired path to the file
    :param contents: The contents to write to the file, should json-serializable
    :return: True if the file was newly created, false otherwise
    """
    try:
        with io.open(path, "w", encoding="utf-8") as f:
            write_to_file_utf8(f, contents)
            f.flush()

    except OSError as e:
        if e.errno == errno.EEXIST:
            # If the file exists no changes necessary
            return False
        if e.errno == errno.EACCES:
            log.error("Cannot create file {}".format(path))
        else:
            raise
    return True


def write_to_file_utf8(file, data: dict) -> None:
    """
    Writes out the provided data to the specified file, with user-friendly indentation,
    and in utf-8 encoding
    :param file: The file to write to
    :param data: The data to be written
    :return:
    """
    file.write(json.dumps(data, indent=2, ensure_ascii=False, sort_keys=True))



#
# Subscription manager command line utility.
#
# Copyright (c) 2021 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#
import os

import rhsm.config

from subscription_manager.cli import system_exit
from subscription_manager.cli_command.cli import CliCommand, conf
from subscription_manager.i18n import ugettext as _


class ConfigCommand(CliCommand):
    def __init__(self):
        shortdesc = _("List, set, or remove the configuration parameters in use by this system")
        super(ConfigCommand, self).__init__("config", shortdesc, False)

        self.parser.add_argument(
            "--list", action="store_true", help=_("list the configuration for this system")
        )
        self.parser.add_argument(
            "--remove", dest="remove", action="append", help=_("remove configuration entry by section.name")
        )
        for s in list(conf.keys()):
            section = conf[s]
            for name, _value in list(section.items()):
                # Allow adding CLI options only for sections and names listed in defaults
                if s in rhsm.config.DEFAULTS and name in rhsm.config.DEFAULTS[s]:
                    self.parser.add_argument(
                        "--" + s + "." + name,
                        dest=(s + "." + name),
                        help=_("Section: {s}, Name: {name}").format(s=s, name=name),
                    )

    def _validate_options(self):
        if self.options.list:
            too_many = False
            if self.options.remove:
                too_many = True
            else:
                for s in list(conf.keys()):
                    section = conf[s]
                    for name, _value in list(section.items()):
                        # Ignore sections and names that are not supported by subscription-manager
                        if hasattr(self.options, s + "." + name):
                            if getattr(self.options, s + "." + name):
                                too_many = True
                                break
                        else:
                            pass
            if too_many:
                system_exit(
                    os.EX_USAGE,
                    _(
                        "Error: --list should not be used with any other "
                        "options for setting or removing configurations."
                    ),
                )

        if not (self.options.list or self.options.remove):
            has = False
            for s in list(conf.keys()):
                section = conf[s]
                for name, _value in list(section.items()):
                    if hasattr(self.options, s + "." + name):
                        test = "{value}".format(value=getattr(self.options, s + "." + name))
                    else:
                        test = "None"
                    has = has or (test != "None")
            if not has:
                # if no options are given, default to --list
                self.options.list = True

        if self.options.remove:
            for r in self.options.remove:
                if "." not in r:
                    system_exit(
                        os.EX_USAGE,
                        _(
                            "Error: configuration entry designation for "
                            "removal must be of format [section.name]"
                        ),
                    )

                section = r.split(".")[0]
                name = r.split(".")[1]
                found = False
                if section in list(conf.keys()):
                    for key, _value in list(conf[section].items()):
                        if name == key:
                            found = True
                if not found:
                    system_exit(
                        os.EX_CONFIG,
                        _("Error: Section {section} and name {name} does not exist.").format(
                            section=section, name=name
                        ),
                    )

    def _do_command(self):
        self._validate_options()

        if self.options.list:
            for s in list(conf.keys()):
                section = conf[s]
                print("[{s}]".format(s=s))
                source_list = sorted(section.items())
                for (name, value) in source_list:
                    indicator1 = ""
                    indicator2 = ""
                    if value == section.get_default(name):
                        indicator1 = "["
                        indicator2 = "]"
                    print(
                        "   {name} = {indicator1}{value}{indicator2}".format(
                            name=name, indicator1=indicator1, value=value, indicator2=indicator2
                        )
                    )
                print()
            print(_("[] - Default value in use"))
            print("\n")
        elif self.options.remove:
            for r in self.options.remove:
                section = r.split(".")[0]
                name = r.split(".")[1]
                try:
                    if not conf[section].has_default(name):
                        conf[section][name] = ""
                        print(
                            _("You have removed the value for section {section} and name {name}.").format(
                                section=section, name=name
                            )
                        )
                    else:
                        conf[section][name] = conf[section].get_default(name)
                        print(
                            _("You have removed the value for section {section} and name {name}.").format(
                                section=section, name=name
                            )
                        )
                        print(_("The default value for {name} will now be used.").format(name=name))
                except Exception:
                    print(
                        _("Section {section} and name {name} cannot be removed.").format(
                            section=section, name=name
                        )
                    )
            conf.persist()
        else:
            for s in list(conf.keys()):
                section = conf[s]
                for name, value in list(section.items()):
                    if hasattr(self.options, s + "." + name):
                        value = "{name}".format(name=getattr(self.options, s + "." + name))
                        if not value == "None":
                            section[name] = value
            conf.persist()

    def require_connection(self):
        return False



# Copyright (c) 2017 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

"""
This module holds information about current state of client application
like sender of D-bus method, current subscription-manager command (register,
attach, ...), dnf command, etc.
"""

import logging
import dbus

import rhsm.utils
from rhsmlib.dbus import dbus_utils

log = logging.getLogger(__name__)


@rhsm.utils.lock
@rhsm.utils.singleton
class DBusSender:
    """
    This class holds information about current sender of D-Bus method
    """

    @rhsm.utils.call_once
    def __init__(self):
        self._cmd_line = None

    @property
    def cmd_line(self):
        with self:
            return self._cmd_line

    @cmd_line.setter
    def cmd_line(self, cmd_line):
        with self:
            self._cmd_line = cmd_line

    @staticmethod
    def get_cmd_line(sender, bus=None):
        """
        Try to get command line of sender
        :param sender: sender
        :param bus: bus
        :return:
        """
        if bus is None:
            bus = dbus.SystemBus()
        cmd_line = dbus_utils.command_of_sender(bus, sender)
        if cmd_line is not None and type(cmd_line) == str:
            # Store only first argument of command line (no argument including username or password)
            cmd_line = cmd_line.split()[0]
        return cmd_line

    def set_cmd_line(self, sender, cmd_line=None, bus=None):
        """
        This method set sender's command line in the singleton object
        :return: None
        """
        if cmd_line is None:
            self.cmd_line = self.get_cmd_line(sender, bus)
        else:
            self.cmd_line = cmd_line
        log.debug("D-Bus sender: %s (cmd-line: %s)" % (sender, self.cmd_line))

    def reset_cmd_line(self):
        """
        Reset sender's command line
        :return: None
        """
        self.cmd_line = None



#!/usr/bin/env python

#
# Copyright (c) 2014 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
import os
import sys
import subprocess

from glob import glob
from setuptools import setup, find_packages, Extension
from setuptools.command.install import install as _install


from distutils import log
from distutils.command.install_data import install_data as _install_data
from distutils.command.build import build as _build
from distutils.command.build_py import build_py as _build_py

# Note that importing build_ext alone won't be enough to make certain tasks (like lint) work
# those tasks require that some dependencies be installed.  Munging the syspath
# here is just so that setup.py will be able to load and run in Jenkins jobs and RPM builds
# that don't set up a proper development environment.
build_ext_home = os.path.abspath(os.path.join(os.path.dirname(__file__), "./build_ext"))
sys.path.append(build_ext_home)
from build_ext import i18n, lint, template, utils  # noqa: E402


# Read packages we should exclude from the environment
# This is used to deal with the fact that we have multiple packages which
# might be built optionally all tracked / installed via one setup.
exclude_packages = [x.strip() for x in os.environ.get("EXCLUDE_PACKAGES", "").split(",") if x != ""]
exclude_packages.extend(
    [
        "*.plugin.ostree",
        "*.services.examples",
    ]
)


PKG_VERSION = None


# subclass build_py so we can generate version.py based on --pkg-version or
# from a guess generated from 'git describe'
class pkg_version_release_build_py(_build_py):
    user_options = _build_py.user_options + [
        ("pkg-version=", None, "version and release of the PKG this is built for")
    ]

    def initialize_options(self):
        _build_py.initialize_options(self)
        self.pkg_version = None
        self.versioned_packages = []

    def finalize_options(self):
        # When --pkg-version was provided as command line
        # option of install command, then get this value from global variables
        if self.pkg_version is None and PKG_VERSION is not None:
            self.pkg_version = PKG_VERSION
        _build_py.finalize_options(self)
        self.set_undefined_options(
            "build",
            ("pkg_version", "pkg_version"),
        )

    def run(self):
        log.debug("Building with PKG_VERSION=%s" % self.pkg_version)
        _build_py.run(self)
        # create a "version.py" that includes the pkg version
        # info passed to our new build_py args
        if not self.dry_run:
            for package in self.versioned_packages:
                version_dir = os.path.join(self.build_lib, package)
                version_file = os.path.join(version_dir, "version.py")
                try:
                    lines = []
                    with open(version_file, "r") as file:
                        for line in file.readlines():
                            line = line.replace("PKG_VERSION", str(self.pkg_version))
                            lines.append(line)

                    with open(version_file, "w") as file:
                        for line in lines:
                            file.write(line)
                except EnvironmentError:
                    raise


class install(_install):
    user_options = _install.user_options + [
        ("pkg-version=", None, "version and release of the PKG this is built for"),
    ]

    def initialize_options(self):
        _install.initialize_options(self)
        self.pkg_version = None

    def finalize_options(self):
        global PKG_VERSION
        if self.pkg_version is not None:
            PKG_VERSION = self.pkg_version
        _install.finalize_options(self)
        self.set_undefined_options(
            "build",
            ("pkg_version", "pkg_version"),
        )


class build(_build):
    user_options = _build.user_options + [
        ("pkg-version=", None, "version and release of the PKG this is built for")
    ]

    def initialize_options(self):
        _build.initialize_options(self)
        self.pkg_version = None
        self.git_tag_prefix = "subscription-manager-"

    def finalize_options(self):
        # When --pkg-version was provided as command line
        # option of install command, then get this value from global variables
        if self.pkg_version is None and PKG_VERSION is not None:
            self.pkg_version = PKG_VERSION

        _build.finalize_options(self)

        # When the pkg-version was not provided as command line options,
        # then try to get such information from .git or rpm
        if not self.pkg_version:
            self.pkg_version = self.get_git_describe()

    def get_git_describe(self):
        try:
            cmd = ["git", "describe"]
            process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
            output = process.communicate()[0].decode("utf-8").strip()
            if output.startswith(self.git_tag_prefix):
                return output[len(self.git_tag_prefix) :]
        except OSError:
            # When building the package there won't be a git repo to introspect so
            # builders *must* specify the version via the --pkg-version option
            return "unknown"

    def has_po_files(self):
        try:
            next(utils.Utils.find_files_of_type("po", "*.po"))
            return True
        except StopIteration:
            return False

    sub_commands = _build.sub_commands + [("build_trans", has_po_files), ("build_template", lambda arg: True)]


class install_data(_install_data):
    """Used to intelligently install data files.  For example, files that must be generated
    (such as .mo files) or an entire tree of data files."""

    user_options = _install_data.user_options + []

    def initialize_options(self):
        _install_data.initialize_options(self)
        # Can't use super() because Command isn't a new-style class.

    def finalize_options(self):
        _install_data.finalize_options(self)

    def run(self):
        self.add_messages()
        self.add_dbus_service_files()
        self.add_systemd_services()
        _install_data.run(self)

    def join(self, *args):
        return os.path.normpath(os.path.join(*args))

    def add_messages(self):
        for lang in os.listdir(self.join("build", "locale")):
            lang_dir = self.join("share", "locale", lang, "LC_MESSAGES")
            lang_file = self.join("build", "locale", lang, "LC_MESSAGES", "rhsm.mo")
            self.data_files.append((lang_dir, [lang_file]))

    def add_dbus_service_files(self):
        """
        Add D-Bus service files to the list of files installed to the system
        """
        dbus_service_directory = self.join("share", "dbus-1", "system-services")
        source_dir = self.join("build", "dbus", "system-services-systemd")
        for template_file in os.listdir(source_dir):
            self.data_files.append((dbus_service_directory, [self.join(source_dir, template_file)]))

    def add_systemd_services(self):
        """
        Add .service files for systemd
        """
        systemd_install_directory = self.join("lib", "systemd", "system")
        source_dir = self.join("build", "dbus", "systemd")
        for file in os.listdir(self.join("build", "dbus", "systemd")):
            self.data_files.append((systemd_install_directory, [self.join(source_dir, file)]))


class GettextWithArgparse(i18n.Gettext):
    def find_py(self):
        # Can't use super since we're descended from a old-style class
        files = i18n.Gettext.find_py(self)

        # We need to grab some strings out of argparse for translation
        import argparse

        argparse_source = "%s.py" % os.path.splitext(argparse.__file__)[0]
        if not os.path.exists(argparse_source):
            raise RuntimeError("Could not find argparse.py at %s" % argparse_source)
        files.append(argparse_source)
        return files


setup_requires = []

install_requires = [
    "iniparse",
    "python-dateutil",
    "ethtool",
    "dbus-python",
]

test_require = (
    [
        "pytest",
        "pytest-randomly",
        "pytest-timeout",
        "coverage",
        "polib",
        "flake8",
    ]
    + install_requires
    + setup_requires
)

cmdclass = {
    "clean": utils.clean,
    "install": install,
    "install_data": install_data,
    "build": build,
    "build_py": pkg_version_release_build_py,
    "build_trans": i18n.BuildTrans,
    "build_template": template.BuildTemplate,
    "update_trans": i18n.UpdateTrans,
    "uniq_trans": i18n.UniqTrans,
    "gettext": GettextWithArgparse,
    "lint": lint.Lint,
    "lint_rpm": lint.RpmLint,
    "flake8": lint.Flake8,
}

setup(
    name="subscription-manager",
    version="1.29.33.1",
    url="http://www.candlepinproject.org",
    description="Manage subscriptions for Red Hat products.",
    license="GPLv2",
    author="Adrian Likins",
    author_email="alikins@redhat.com",
    cmdclass=cmdclass,
    packages=find_packages("src", exclude=exclude_packages),
    package_dir={"": "src"},
    entry_points={
        "console_scripts": [
            "subscription-manager = subscription_manager.scripts.subscription_manager:main",
            "rct = subscription_manager.scripts.rct:main",
            "rhsm-debug = subscription_manager.scripts.rhsm_debug:main",
            "rhsm-facts-service = subscription_manager.scripts.rhsm_facts_service:main",
            "rhsm-service = subscription_manager.scripts.rhsm_service:main",
            "rhsmcertd-worker = subscription_manager.scripts.rhsmcertd_worker:main",
            "rhsm-package-profile-uploader = subscription_manager.scripts.package_profile_upload:main",
        ],
    },
    data_files=[
        # man pages for gui are added in add_gui_doc_files(), when GUI package is created
        ("share/man/man8", set(glob("man/*.8"))),
        ("share/man/man5", glob("man/*.5")),
    ],
    command_options={
        "egg_info": {
            "egg_base": ("setup.py", os.curdir),
        },
        "build_py": {
            "versioned_packages": ("setup.py", ["subscription_manager", "rct"]),
        },
    },
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GPL License",
        "Operating System :: Linux",
    ],
    include_package_data=True,
    setup_requires=setup_requires,
    install_requires=install_requires,
    tests_require=test_require,
    ext_modules=[
        Extension("rhsm._certificate", ["src/certificate.c"], libraries=["ssl", "crypto"]),
    ],
)



# Copyright (c) 2015 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#

"""
This file contains code of dnf subcommand called "upload" used for uploading of combined profile (list of rpm
packages, enabled repositories, modules).
"""


from dnfpluginscore import logger
import dnf.cli

from subscription_manager import packageprofilelib
from subscription_manager.injectioninit import init_dep_injection
from subscription_manager.i18n import ugettext as _


@dnf.plugin.register_command
class UploadProfileCommand(dnf.cli.Command):
    name = "upload-profile"
    aliases = (
        "uploadprofile",
        "upload-profile",
    )
    summary = _(
        "Upload combined profile to Satellite server (list of installed rpms, "
        "enabled repositories and modules"
    )

    def __init__(self, cli):
        super(UploadProfileCommand, self).__init__(cli)

    @staticmethod
    def set_argparser(parser):
        parser.add_argument(
            "--force-upload",
            action="store_true",
            help=_("Force package profile upload"),
        )

    def configure(self):
        pass

    def run(self):
        try:
            init_dep_injection()
        except ImportError as e:
            logger.error(str(e))
            return

        command = packageprofilelib.PackageProfileActionCommand()
        report = command.perform(force_upload=self.opts.force_upload)

        if report._status == 0:
            print(_("No updates performed. See /var/log/rhsm/rhsm.log for more information."))
        else:
            print(report)



# Copyright (c) 2016 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.

import logging
import sys
import decorator
import dbus.service
import json
import re

from rhsmlib.dbus import exceptions
from rhsmlib.client_info import DBusSender

log = logging.getLogger(__name__)

__all__ = [
    "dbus_handle_exceptions",
    "dbus_handle_sender",
    "dbus_service_method",
    "dbus_service_signal",
]


@decorator.decorator
def dbus_handle_sender(func, *args, **kwargs):
    """
    Decorator to handle sender argument
    :param func: method with implementation of own logic of D-Bus method
    :param args: arguments of D-Bus method
    :param kwargs: keyed arguments of D-Bus method
    :return: result of D-Bus method
    """

    sender = None
    # Get sender from arguments
    if "sender" in kwargs:
        sender = kwargs["sender"]
    elif len(args) > 0:
        sender = args[-1]

    with DBusSender() as dbus_sender:
        if sender is not None:
            dbus_sender.set_cmd_line(sender)

        try:
            return func(*args, **kwargs)
        finally:
            if sender is not None:
                # When sender was specified, then reset it
                dbus_sender.reset_cmd_line()


@decorator.decorator
def dbus_handle_exceptions(func, *args, **kwargs):
    """
    Decorator to handle exceptions, log them, and wrap them if necessary.
    """

    try:
        return func(*args, **kwargs)
    except Exception as err:
        log.exception(err)
        trace = sys.exc_info()[2]

        severity = "error"
        # Remove "HTTP error (...): " string from the messages:
        pattern = "^HTTP error \x28.*\x29: "
        err_msg = re.sub(pattern, "", str(err))
        # Modify severity of some exception here
        if "Ignoring request to auto-attach. It is disabled for org" in err_msg:
            severity = "warning"
        if hasattr(err, "severity"):
            severity = err.severity
        # Raise exception string as JSON string. Thus it can be parsed and printed properly.
        error_msg = json.dumps(
            {
                "exception": type(err).__name__,
                "severity": severity,
                "message": err_msg,
            }
        )
        raise exceptions.RHSM1DBusException(error_msg).with_traceback(trace)


def dbus_service_method(*args, **kwargs):
    # Tell python-dbus that "sender" will be the keyword to use for the sender unless otherwise
    # defined.
    kwargs.setdefault("sender_keyword", "sender")
    return dbus.service.method(*args, **kwargs)


def dbus_service_signal(*args, **kwargs):
    """
    Decorator used for signal
    :param args:
    :param kwargs:
    :return:
    """
    return dbus.service.signal(*args, **kwargs)



# Copyright (c) 2019 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
#

import logging
import json
from typing import Any, Dict, Callable, List, Optional, Union

from cloud_what._base_provider import BaseCloudProvider

from cloud_what.provider import get_cloud_provider, DetectionMethod
from rhsmlib.facts import collector


log = logging.getLogger(__name__)


class CloudFactsCollector(collector.FactsCollector):
    """
    Class used for collecting facts related to Cloud instances
    """

    def __init__(
        self,
        arch: str = None,
        prefix: str = None,
        testing: bool = None,
        collected_hw_info: Dict[str, Union[str, None]] = None,
    ):
        super(CloudFactsCollector, self).__init__(
            arch=arch,
            prefix=prefix,
            testing=testing,
            collected_hw_info=collected_hw_info,
        )

        self.hardware_methods: List[Callable] = []

        # Try to detect cloud provider using only strong method
        self.cloud_provider: Optional[BaseCloudProvider] = get_cloud_provider(
            facts=self._collected_hw_info,
            methods=DetectionMethod.STRONG,
        )

        if self.cloud_provider is not None:
            # Create dispatcher for supported cloud providers
            cloud_provider_dispatcher: Dict[str, Callable] = {
                "aws": self.get_aws_facts,
                "azure": self.get_azure_facts,
                "gcp": self.get_gcp_facts,
            }
            # Set method according detected cloud provider
            if self.cloud_provider.CLOUD_PROVIDER_ID in cloud_provider_dispatcher:
                self.hardware_methods: List[Callable] = [
                    cloud_provider_dispatcher[self.cloud_provider.CLOUD_PROVIDER_ID]
                ]

    def get_aws_facts(self) -> Dict[str, Union[str, None]]:
        """
        Try to get AWS facts (only instance ID ATM) of machine running on AWS public cloud
        :return:
            dictionary containing {"aws_instance_id": some_instance_ID} when
            the machine is able to gather metadata from AWS cloud provider;
            otherwise returns empty dictionary {}
        """

        metadata_str: str = self.cloud_provider.get_metadata()

        facts: Dict[str, Union[str, None]] = {}
        if metadata_str is not None:
            values: dict[str, Union[str, None]] = self.parse_json_content(metadata_str)

            # Add these three attributes to system facts
            if "instanceId" in values:
                facts["aws_instance_id"] = values["instanceId"]

            if "accountId" in values:
                facts["aws_account_id"] = values["accountId"]

            # BTW: There should be only two types of billing codes: bp-63a5400a and bp-6fa54006 in the list,
            # when RHEL is used. When the subscription-manager is used by some other Linux distribution,
            # then there could be different codes or it could be null
            if "billingProducts" in values:
                billing_products: Optional[List[str]] = values["billingProducts"]
                if isinstance(billing_products, list):
                    facts["aws_billing_products"] = " ".join(billing_products)
                elif billing_products is None:
                    facts["aws_billing_products"] = billing_products
                else:
                    log.debug("AWS metadata attribute billingProducts has to be list or null")

            if "marketplaceProductCodes" in values:
                marketplace_product_codes = values["marketplaceProductCodes"]
                if isinstance(marketplace_product_codes, list):
                    facts["aws_marketplace_product_codes"] = " ".join(marketplace_product_codes)
                elif marketplace_product_codes is None:
                    facts["aws_marketplace_product_codes"] = marketplace_product_codes
                else:
                    log.debug("AWS metadata attribute marketplaceProductCodes has to be list or null")

        return facts

    def get_azure_facts(self) -> Dict[str, str]:
        """
        Try to get facts of VM running on Azure public cloud. Returned dictionary has following format:
            {
                "azure_instance_id": some_instance_ID,
                "azure_offer": some_offer,
                "azure_sku": some_sku
            }
        :return: dictionary containing Azure facts, when the machine is able to gather metadata
            from Azure cloud provider; otherwise returns empty dictionary {}
        """

        metadata_str: str = self.cloud_provider.get_metadata()

        facts: Dict[str, str] = {}
        if metadata_str is not None:
            values: Dict[str, Any] = self.parse_json_content(metadata_str)
            if "compute" in values:
                if "vmId" in values["compute"]:
                    facts["azure_instance_id"] = values["compute"]["vmId"]
                if "sku" in values["compute"]:
                    facts["azure_sku"] = values["compute"]["sku"]
                if "offer" in values["compute"]:
                    facts["azure_offer"] = values["compute"]["offer"]
        return facts

    def get_gcp_facts(self) -> Dict[str, str]:
        """
        Try to get facts of VM running on GCP public cloud. Only instance_id is reported ATM.
        :return: dictionary containing GCP facts, when the machine is able to gather metadata
            from GCP cloud provider; otherwise returns empty dictionary {}
        """

        encoded_jwt_token: str = self.cloud_provider.get_metadata()

        facts: Dict[str, str] = {}
        if encoded_jwt_token is not None:
            jose_header, metadata, signature = self.cloud_provider.decode_jwt(encoded_jwt_token)
            if metadata is not None:
                values: Dict[str, Any] = self.parse_json_content(metadata)
                if "google" in values and "compute_engine" in values["google"]:
                    # ID of instance
                    if "instance_id" in values["google"]["compute_engine"]:
                        facts["gcp_instance_id"] = values["google"]["compute_engine"]["instance_id"]
                    else:
                        log.debug("GCP instance_id not found in JWT token")
                    # IDs of licenses
                    if "license_id" in values["google"]["compute_engine"]:
                        gcp_license_codes = values["google"]["compute_engine"]["license_id"]
                        facts["gcp_license_codes"] = " ".join(gcp_license_codes)
                    else:
                        log.debug("GCP license codes not found in JWT token")
                else:
                    log.debug("GCP google.compute_engine on found in JWT token")
        return facts

    @staticmethod
    def parse_json_content(content: str) -> Dict[str, Any]:
        """
        Parse content returned from AWS metadata provider
        :param content: string of JSON document
        :return: Dictionary containing values from parsed JSON document
        """
        try:
            return json.loads(content)
        except ValueError as e:
            raise ValueError("Failed to parse json data with error: %s", str(e))



# Copyright (c) 2016 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
import ast

from distutils.spawn import spawn

from build_ext.utils import Utils, BaseCommand
import os
import subprocess
from shutil import which

# These dependencies aren't available in build environments.  We won't need any
# linting functionality there though, so just create a dummy class so we can proceed.
try:
    import pycodestyle
except ImportError:
    pycodestyle = None


class Lint(BaseCommand):
    description = "examine code for errors"

    def has_spec_file(self):
        try:
            next(Utils.find_files_of_type(".", "*.spec"))
            return True
        except StopIteration:
            return False

    def has_flake8(self):
        return which("flake8") is not None

    def run(self):
        for cmd_name in self.get_sub_commands():
            self.run_command(cmd_name)

    # Defined at the end since it references unbound methods
    sub_commands = [
        ("lint_rpm", has_spec_file),
        ("flake8", has_flake8),
    ]


class RpmLint(BaseCommand):
    description = "run rpmlint on spec files"

    def run(self):
        files = subprocess.run(["git", "ls-files", "--full-name"], capture_output=True).stdout
        files = files.decode().splitlines()
        files = [x for x in files if x.endswith(".spec")]
        for f in files:
            spawn(["rpmlint", os.path.realpath(f)])


class Flake8(BaseCommand):
    description = "run flake8"

    def run(self):
        spawn(["flake8"])


class AstVisitor(object):
    """Visitor pattern for looking at specific nodes in an AST.  Basically a copy of
    ast.NodeVisitor, but with the additional feature of appending return values onto a result
    list that is ultimately returned.

    I recommend reading http://greentreesnakes.readthedocs.io/en/latest/index.html for a good
    overview of the various Python AST node types.
    """

    def __init__(self):
        self.results = []

    def visit(self, node):
        method = "visit_" + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        r = visitor(node)
        if r is not None:
            self.results.append(r)
        return self.results

    def generic_visit(self, node):
        for field, value in ast.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        self.visit(item)
            elif isinstance(value, ast.AST):
                self.visit(value)


class DebugImportVisitor(AstVisitor):
    """Look for imports of various debug modules"""

    DEBUG_MODULES = ["pdb", "pudb", "ipdb", "pydevd"]
    codes = ["X200"]

    def visit_Import(self, node):
        # Likely not necessary but prudent
        self.generic_visit(node)

        for alias in node.names:
            module_name = alias.name
            if module_name in self.DEBUG_MODULES:
                return (node, "X200 imports of debug module '%s' should be removed" % module_name)

    def visit_ImportFrom(self, node):
        # Likely not necessary but prudent
        self.generic_visit(node)
        module_name = node.module
        if module_name in self.DEBUG_MODULES:
            return (node, "X200 imports of debug module '%s' should be removed" % module_name)


class GettextVisitor(AstVisitor):
    """Looks for Python string formats that are known to break xgettext.
    Specifically, constructs of the forms:
        _("a" + "b")
        _("a" + \
        "b")
    Also look for _(a) usages
    """

    codes = ["X300", "X301", "X302"]

    def visit_Call(self, node):
        # Descend first
        self.generic_visit(node)

        func = node.func
        if not isinstance(func, ast.Name):
            return

        if func.id != "_":
            return

        for arg in node.args:
            # ProTip: use print(ast.dump(node)) to figure out what the node looks like

            # Things like _("a" + "b") (including such constructs across line continuations
            if isinstance(arg, ast.BinOp) and isinstance(arg.op, ast.Add):
                return (node, "X300 string concatenation that will break xgettext")

            # Things like _(some_variable)
            if isinstance(arg, ast.Name):
                return (node, "X301 variable reference that will break xgettext")

            # _("%s is great" % some_variable) should be _("%s is great") % some_variable
            if isinstance(arg, ast.BinOp) and isinstance(arg.op, ast.Mod):
                return (
                    node,
                    "X302 string formatting within gettext function: _('%s' % foo) should be _('%s') % foo",
                )


class AstChecker(object):
    name = "SubscriptionManagerAstChecker"
    version = "1.0"

    def __init__(self, tree, filename):
        self.tree = tree
        self.filename = filename

        self.visitors = [
            (GettextVisitor, {}),
            (DebugImportVisitor, {}),
        ]

    def run(self):
        if self.tree:
            for visitor, kwargs in self.visitors:
                result = visitor(**kwargs).visit(self.tree)
                if result:
                    for node, msg in result:
                        yield self.err(node, msg)

    def err(self, node, msg=None):
        if not msg:
            msg = self._error_tmpl

        lineno = getattr(node, "lineno", 1)
        col_offset = getattr(node, "col_offset", 0)

        # Adjust line number and offset if a decorator is applied
        if isinstance(node, ast.ClassDef):
            lineno += len(node.decorator_list)
            col_offset += 6
        elif isinstance(node, ast.FunctionDef):
            lineno += len(node.decorator_list)
            col_offset += 4

        ret = (lineno, col_offset, msg, self)
        return ret



import unittest

import locale

from datetime import datetime
from dateutil.tz import tzutc, tzstr
from unittest.mock import patch

from subscription_manager import managerlib


class TestFormatTime(unittest.TestCase):
    def setUp(self):
        self.locale = locale.getlocale()
        locale.setlocale(locale.LC_ALL, "en_US")

    def tearDown(self):
        locale.setlocale(locale.LC_ALL, self.locale)

    @patch("subscription_manager.managerlib.tzlocal")
    def test_system_dst(self, mock_tz):
        mock_tz.return_value = tzstr("EST5EDT")
        test = datetime(2014, 12, 21, 4, 59, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "12/20/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-12-21")
        test = datetime(2014, 12, 21, 5, 0, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "12/21/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-12-21")
        test = datetime(2014, 12, 21, 5, 1, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "12/21/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-12-21")
        test = datetime(2014, 12, 21, 3, 49, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "12/20/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-12-21")
        test = datetime(2014, 12, 21, 4, 0, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "12/20/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-12-21")
        test = datetime(2014, 12, 21, 4, 1, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "12/20/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-12-21")

        test = datetime(2014, 5, 21, 4, 59, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "05/21/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-05-21")
        test = datetime(2014, 5, 21, 5, 0, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "05/21/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-05-21")
        test = datetime(2014, 5, 21, 5, 1, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "05/21/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-05-21")
        test = datetime(2014, 5, 21, 3, 59, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "05/20/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-05-21")
        test = datetime(2014, 5, 21, 4, 0, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "05/21/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-05-21")
        test = datetime(2014, 5, 21, 4, 1, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "05/21/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-05-21")

    @patch("subscription_manager.managerlib.tzlocal")
    def test_system_est(self, mock_tz):
        mock_tz.return_value = tzstr("EST5EDT")
        test = datetime(2014, 12, 21, 4, 59, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "12/20/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-12-21")
        test = datetime(2014, 12, 21, 5, 0, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "12/21/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-12-21")
        test = datetime(2014, 12, 21, 5, 1, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "12/21/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-12-21")
        test = datetime(2014, 12, 21, 3, 49, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "12/20/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-12-21")
        test = datetime(2014, 12, 21, 4, 0, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "12/20/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-12-21")
        test = datetime(2014, 12, 21, 4, 1, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "12/20/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-12-21")

        test = datetime(2014, 5, 21, 4, 59, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "05/21/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-05-21")
        test = datetime(2014, 5, 21, 5, 0, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "05/21/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-05-21")
        test = datetime(2014, 5, 21, 5, 1, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "05/21/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-05-21")
        test = datetime(2014, 5, 21, 3, 59, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "05/20/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-05-21")
        test = datetime(2014, 5, 21, 4, 0, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "05/21/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-05-21")
        test = datetime(2014, 5, 21, 4, 1, 0, tzinfo=tzutc())
        self.assertEqual(managerlib.format_date(test), "05/21/2014")
        self.assertEqual(managerlib.format_iso8601_date(test), "2014-05-21")



# Copyright (c) 2022 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public
# License as published by the Free Software Foundation; either version
# 2 of the License (GPLv2) or (at your option) any later version.
# There is NO WARRANTY for this software, express or implied,
# including the implied warranties of MERCHANTABILITY,
# NON-INFRINGEMENT, or FITNESS FOR A PARTICULAR PURPOSE. You should
# have received a copy of GPLv2 along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#

from rhsmlib.facts.dmidecodeparser import DmidecodeParser
from rhsmlib.facts.dmiinfo import DmidecodeFactCollector

import contextlib
import os
import unittest


FAKE_PART_NUMBER = "AAAAAAAAAAAAA-ZZ"
FAKE_PRODUCT_NAME = "PPPPPPPPPP"
FAKE_SERIAL_NUMBER = "SSSSSSSS"
FAKE_UUID = "11111111-2222-3333-4444-555555555555"


class DmidecodeTestDataMixin:
    """
    Simple mixin for testing dmidecode output files.
    """

    DATA_SUFFIX = ".txt"

    def __init__(self, *args, **kwargs):
        curdir = os.path.dirname(os.path.realpath(__file__))
        self._datadir = os.path.join(curdir, "dmidecodedata")
        super(DmidecodeTestDataMixin, self).__init__(*args, **kwargs)

    @property
    def datadir(self):
        return self._datadir

    def load_data(self, name):
        name = os.path.join(self.datadir, name)
        parser = DmidecodeParser()
        parser.parse_file(name)
        return parser

    def get_testfiles(self):
        testfiles = []
        with os.scandir(self.datadir) as it:
            for entry in it:
                if not entry.name.endswith(self.DATA_SUFFIX) or not entry.is_file():
                    continue
                testfiles.append(entry.name)
        return testfiles


class TestDmidecodeParser(DmidecodeTestDataMixin, unittest.TestCase):
    def test_data(self):
        testfiles = self.get_testfiles()

        for tf in testfiles:
            with self.subTest(file=tf):
                parser = self.load_data(tf)
                self.assertEqual(
                    parser.get_key(DmidecodeParser.DmiTypes.SYSTEM_INFORMATION, "UUID"), FAKE_UUID
                )
                with contextlib.suppress(KeyError):
                    # not all the systems have a serial number set
                    self.assertEqual(
                        parser.get_key(DmidecodeParser.DmiTypes.SYSTEM_INFORMATION, "Serial Number"),
                        FAKE_SERIAL_NUMBER,
                    )


class TestDmidecodeFactCollector(DmidecodeTestDataMixin, unittest.TestCase):
    # subtags of "dmi" that we expected to be present in the facts
    FACTS_SUBTAGS = [
        "dmi.bios.",
        "dmi.chassis.",
        "dmi.system.",
    ]

    def test_data(self):
        datadir = self.datadir
        testfiles = self.get_testfiles()

        for tf in testfiles:
            with self.subTest(file=tf):
                c = DmidecodeFactCollector()
                c.set_dmidecode_output(os.path.join(datadir, tf))
                facts = c.get_all()
                for subtag in self.FACTS_SUBTAGS:
                    with self.subTest(subtag=subtag):
                        # at least one fact for the specific subtag
                        self.assertTrue(any(f.startswith(subtag) for f in facts))
                self.assertEqual(facts["dmi.system.uuid"], FAKE_UUID)
                with contextlib.suppress(KeyError):
                    # not all the systems have a serial number set
                    self.assertEqual(facts["dmi.system.serial_number"], FAKE_SERIAL_NUMBER)



# Copyright (c) 2016 Red Hat, Inc.
#
# This software is licensed to you under the GNU General Public License,
# version 2 (GPLv2). There is NO WARRANTY for this software, express or
# implied, including the implied warranties of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. You should have received a copy of GPLv2
# along with this software; if not, see
# http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt.
#
# Red Hat trademarks are not licensed under GPLv2. No permission is
# granted to use or replicate Red Hat trademarks that are incorporated
# in this software or its documentation.
import contextlib
import json
import tempfile
from typing import Optional

from unittest import mock
import socket
import dbus


from rhsm import connection

import rhsmlib.dbus.exceptions
from rhsmlib.dbus.server import DomainSocketServer
from rhsmlib.dbus.objects import RegisterDBusObject, DomainSocketRegisterDBusObject

from test.rhsmlib.base import DBusServerStubProvider


CONSUMER_CONTENT_JSON = """{"hypervisorId": "foo",
        "serviceLevel": "",
        "autoheal": true,
        "idCert": {
          "key": "FAKE_KEY",
          "cert": "FAKE_CERT",
          "serial" : {
            "id" : 5196045143213189102,
            "revoked" : false,
            "collected" : false,
            "expiration" : "2033-04-25T18:03:06+0000",
            "serial" : 5196045143213189102,
            "created" : "2017-04-25T18:03:06+0000",
            "updated" : "2017-04-25T18:03:06+0000"
          },
          "id" : "8a8d011e5ba64700015ba647fbd20b88",
          "created" : "2017-04-25T18:03:07+0000",
          "updated" : "2017-04-25T18:03:07+0000"
        },
        "owner": {
          "href": "/owners/admin",
          "displayName": "Admin Owner",
          "id": "ff808081550d997c01550d9adaf40003",
          "key": "admin",
          "contentAccessMode": "entitlement"
        },
        "href": "/consumers/c1b8648c-6f0a-4aa5-b34e-b9e62c0e4364",
        "facts": {}, "id": "ff808081550d997c015511b0406d1065",
        "uuid": "c1b8648c-6f0a-4aa5-b34e-b9e62c0e4364",
        "guestIds": null, "capabilities": null,
        "environment": null, "installedProducts": null,
        "canActivate": false, "type": {"manifest": false,
        "id": "1000", "label": "system"}, "annotations": null,
        "username": "admin", "updated": "2016-06-02T15:16:51+0000",
        "lastCheckin": null, "entitlementCount": 0, "releaseVer":
        {"releaseVer": null}, "entitlementStatus": "valid", "name":
        "test.example.com", "created": "2016-06-02T15:16:51+0000",
        "contentTags": null, "dev": false}"""

CONSUMER_CONTENT_JSON_SCA = """{"hypervisorId": null,
        "serviceLevel": "",
        "autoheal": true,
        "idCert": {
          "key": "FAKE_KEY",
          "cert": "FAKE_CERT",
          "serial" : {
            "id" : 5196045143213189102,
            "revoked" : false,
            "collected" : false,
            "expiration" : "2033-04-25T18:03:06+0000",
            "serial" : 5196045143213189102,
            "created" : "2017-04-25T18:03:06+0000",
            "updated" : "2017-04-25T18:03:06+0000"
          },
          "id" : "8a8d011e5ba64700015ba647fbd20b88",
          "created" : "2017-04-25T18:03:07+0000",
          "updated" : "2017-04-25T18:03:07+0000"
        },
        "owner": {
          "href": "/owners/admin",
          "displayName": "Admin Owner",
          "id": "ff808081550d997c01550d9adaf40003",
          "key": "admin",
          "contentAccessMode": "org_environment"
        },
        "href": "/consumers/c1b8648c-6f0a-4aa5-b34e-b9e62c0e4364",
        "facts": {}, "id": "ff808081550d997c015511b0406d1065",
        "uuid": "c1b8648c-6f0a-4aa5-b34e-b9e62c0e4364",
        "guestIds": null, "capabilities": null,
        "environment": null, "installedProducts": null,
        "canActivate": false, "type": {"manifest": false,
        "id": "1000", "label": "system"}, "annotations": null,
        "username": "admin", "updated": "2016-06-02T15:16:51+0000",
        "lastCheckin": null, "entitlementCount": 0, "releaseVer":
        {"releaseVer": null}, "entitlementStatus": "valid", "name":
        "test.example.com", "created": "2016-06-02T15:16:51+0000",
        "contentTags": null, "dev": false}"""

ENABLED_CONTENT = """[ {
  "created" : "2022-06-30T13:24:33+0000",
  "updated" : "2022-06-30T13:24:33+0000",
  "id" : "16def3d98d6549f8a3649f723a76991c",
  "consumer" : {
    "id" : "4028face81aa047e0181b4c8e1170bdc",
    "uuid" : "8503a41a-6ce2-480c-bc38-b67d6aa6dd20",
    "name" : "thinkpad-t580",
    "href" : "/consumers/8503a41a-6ce2-480c-bc38-b67d6aa6dd20"
  },
  "pool" : {
    "created" : "2022-06-28T11:14:35+0000",
    "updated" : "2022-06-30T13:24:33+0000",
    "id" : "4028face81aa047e0181aa052f740360",
    "type" : "NORMAL",
    "owner" : {
      "id" : "4028face81aa047e0181aa0490e30002",
      "key" : "admin",
      "displayName" : "Admin Owner",
      "href" : "/owners/admin",
      "contentAccessMode" : "entitlement"
    },
    "activeSubscription" : true,
    "sourceEntitlement" : null,
    "quantity" : 5,
    "startDate" : "2022-06-23T13:14:26+0000",
    "endDate" : "2023-06-23T13:14:26+0000",
    "attributes" : [ ],
    "restrictedToUsername" : null,
    "contractNumber" : "0",
    "accountNumber" : "6547096716",
    "orderNumber" : "order-23226139",
    "consumed" : 1,
    "exported" : 0,
    "branding" : [ ],
    "calculatedAttributes" : {
      "compliance_type" : "Standard"
    },
    "upstreamPoolId" : "upstream-05736148",
    "upstreamEntitlementId" : null,
    "upstreamConsumerId" : null,
    "productName" : "SP Server Standard (U: Development, R: SP Server)",
    "productId" : "sp-server-dev",
    "productAttributes" : [ {
      "name" : "management_enabled",
      "value" : "1"
    }, {
      "name" : "usage",
      "value" : "Development"
    }, {
      "name" : "roles",
      "value" : "SP Server"
    }, {
      "name" : "variant",
      "value" : "ALL"
    }, {
      "name" : "sockets",
      "value" : "128"
    }, {
      "name" : "support_level",
      "value" : "Standard"
    }, {
      "name" : "support_type",
      "value" : "L1-L3"
    }, {
      "name" : "arch",
      "value" : "ALL"
    }, {
      "name" : "type",
      "value" : "MKT"
    }, {
      "name" : "version",
      "value" : "1.0"
    } ],
    "stackId" : null,
    "stacked" : false,
    "sourceStackId" : null,
    "developmentPool" : false,
    "href" : "/pools/4028face81aa047e0181aa052f740360",
    "derivedProductAttributes" : [ ],
    "derivedProductId" : null,
    "derivedProductName" : null,
    "providedProducts" : [ {
      "productId" : "99000",
      "productName" : "SP Server Bits"
    } ],
    "derivedProvidedProducts" : [ ],
    "subscriptionSubKey" : "master",
    "subscriptionId" : "srcsub-45255972",
    "locked" : false
  },
  "quantity" : 1,
  "certificates" : [ {
    "created" : "2022-06-30T13:24:33+0000",
    "updated" : "2022-06-30T13:24:33+0000",
    "id" : "4028face81aa047e0181b4c8e4b90be1",
    "key" : "-----BEGIN PRIVATE KEY-----REDACTED-----END PRIVATE KEY-----",
    "cert" : "-----BEGIN CERTIFICATE-----REDACTED-----END RSA SIGNATURE-----",
    "serial" : {
      "created" : "2022-06-30T13:24:33+0000",
      "updated" : "2022-06-30T13:24:33+0000",
      "id" : 3712610178651551557,
      "serial" : 3712610178651551557,
      "expiration" : "2023-06-23T13:14:26+0000",
      "revoked" : false
    }
  } ],
  "startDate" : "2022-06-23T13:14:26+0000",
  "endDate" : "2023-06-23T13:14:26+0000",
  "href" : null
} ]
"""

# Following consumer do not contain information about content access mode
OLD_CONSUMER_CONTENT_JSON = """{"hypervisorId": null,
        "serviceLevel": "",
        "autoheal": true,
        "idCert": {
          "key": "FAKE_KEY",
          "cert": "FAKE_CERT",
          "serial" : {
            "id" : 5196045143213189102,
            "revoked" : false,
            "collected" : false,
            "expiration" : "2033-04-25T18:03:06+0000",
            "serial" : 5196045143213189102,
            "created" : "2017-04-25T18:03:06+0000",
            "updated" : "2017-04-25T18:03:06+0000"
          },
          "id" : "8a8d011e5ba64700015ba647fbd20b88",
          "created" : "2017-04-25T18:03:07+0000",
          "updated" : "2017-04-25T18:03:07+0000"
        },
        "owner": {
          "href": "/owners/admin",
          "displayName": "Admin Owner",
          "id": "ff808081550d997c01550d9adaf40003",
          "key": "admin"
        },
        "href": "/consumers/c1b8648c-6f0a-4aa5-b34e-b9e62c0e4364",
        "facts": {}, "id": "ff808081550d997c015511b0406d1065",
        "uuid": "c1b8648c-6f0a-4aa5-b34e-b9e62c0e4364",
        "guestIds": null, "capabilities": null,
        "environments": null, "installedProducts": null,
        "canActivate": false, "type": {"manifest": false,
        "id": "1000", "label": "system"}, "annotations": null,
        "username": "admin", "updated": "2016-06-02T15:16:51+0000",
        "lastCheckin": null, "entitlementCount": 0, "releaseVer":
        {"releaseVer": null}, "entitlementStatus": "valid", "name":
        "test.example.com", "created": "2016-06-02T15:16:51+0000",
        "contentTags": null, "dev": false}"""

OWNERS_CONTENT_JSON = """[
    {
        "autobindDisabled": false,
        "autobindHypervisorDisabled": false,
        "contentAccessMode": "entitlement",
        "contentAccessModeList": "entitlement",
        "contentPrefix": null,
        "created": "2020-02-17T08:21:47+0000",
        "defaultServiceLevel": null,
        "displayName": "Donald Duck",
        "href": "/owners/donaldduck",
        "id": "ff80808170523d030170523d34890003",
        "key": "donaldduck",
        "lastRefreshed": null,
        "logLevel": null,
        "parentOwner": null,
        "updated": "2020-02-17T08:21:47+0000",
        "upstreamConsumer": null
    },
    {
        "autobindDisabled": false,
        "autobindHypervisorDisabled": false,
        "contentAccessMode": "entitlement",
        "contentAccessModeList": "entitlement",
        "contentPrefix": null,
        "created": "2020-02-17T08:21:47+0000",
        "defaultServiceLevel": null,
        "displayName": "Admin Owner",
        "href": "/owners/admin",
        "id": "ff80808170523d030170523d347c0002",
        "key": "admin",
        "lastRefreshed": null,
        "logLevel": null,
        "parentOwner": null,
        "updated": "2020-02-17T08:21:47+0000",
        "upstreamConsumer": null
    },
    {
        "autobindDisabled": false,
        "autobindHypervisorDisabled": false,
        "contentAccessMode": "entitlement",
        "contentAccessModeList": "entitlement,org_environment",
        "contentPrefix": null,
        "created": "2020-02-17T08:21:47+0000",
        "defaultServiceLevel": null,
        "displayName": "Snow White",
        "href": "/owners/snowwhite",
        "id": "ff80808170523d030170523d348a0004",
        "key": "snowwhite",
        "lastRefreshed": null,
        "logLevel": null,
        "parentOwner": null,
        "updated": "2020-02-17T08:21:47+0000",
        "upstreamConsumer": null
    }
]
"""


class RegisterDBusObjectTest(DBusServerStubProvider):
    dbus_class = RegisterDBusObject
    dbus_class_kwargs = {}
    socket_dir: Optional[tempfile.TemporaryDirectory] = None

    def setUp(self) -> None:
        self.socket_dir = tempfile.TemporaryDirectory()
        self.addCleanup(self.socket_dir.cleanup)

        socket_path_patch = mock.patch.object(DomainSocketServer, "_server_socket_path", self.socket_dir.name)
        socket_path_patch.start()
        # `tmpdir` behaves differently from `dir` on old versions of dbus
        # (earlier than 1.12.24 and 1.14.4).
        # In newer versions we are not getting abstract socket anymore.
        socket_iface_patch = mock.patch.object(DomainSocketServer, "_server_socket_iface", "unix:dir=")
        socket_iface_patch.start()

        super().setUp()

    def tearDown(self) -> None:
        """Make sure the domain server is stopped once the test ends."""
        with contextlib.suppress(rhsmlib.dbus.exceptions.Failed):
            self.obj.Stop.__wrapped__(self.obj, self.LOCALE)

        super().tearDown()

    def test_Start(self):
        substring = self.socket_dir.name + "/dbus.*"
        result = self.obj.Start.__wrapped__(self.obj, self.LOCALE)
        self.assertRegex(result, substring)

    def test_Start__two_starts(self):
        """Test that opening the server twice returns the same address"""
        result_1 = self.obj.Start.__wrapped__(self.obj, self.LOCALE)
        result_2 = self.obj.Start.__wrapped__(self.obj, self.LOCALE)
        self.assertEqual(result_1, result_2)

    def test_Start__can_connect(self):
        result = self.obj.Start.__wrapped__(self.obj, self.LOCALE)
        prefix, _, data = result.partition("=")
        address, _, guid = data.partition(",")

        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        try:
            sock.connect(address)
        finally:
            sock.close()

    def test_Stop(self):
        self.obj.Start.__wrapped__(self.obj, self.LOCALE)
        self.obj.Stop.__wrapped__(self.obj, self.LOCALE)

    def test_Stop__not_running(self):
        with self.assertRaises(rhsmlib.dbus.exceptions.Failed):
            self.obj.Stop.__wrapped__(self.obj, self.LOCALE)


class DomainSocketRegisterDBusObjectTest(DBusServerStubProvider):
    dbus_class = DomainSocketRegisterDBusObject
    dbus_class_kwargs = {}

    @classmethod
    def setUpClass(cls) -> None:
        register_patch = mock.patch(
            "rhsmlib.dbus.objects.register.RegisterService.register",
            name="register",
        )
        cls.patches["register"] = register_patch.start()
        cls.addClassCleanup(register_patch.stop)

        unregister_patch = mock.patch(
            "rhsmlib.services.unregister.UnregisterService.unregister", name="unregister"
        )
        cls.patches["unregister"] = unregister_patch.start()
        cls.addClassCleanup(unregister_patch.stop)

        is_registered_patch = mock.patch(
            "rhsmlib.dbus.base_object.BaseObject.is_registered",
            name="is_registered",
        )
        cls.patches["is_registered"] = is_registered_patch.start()
        cls.addClassCleanup(is_registered_patch.stop)

        update_patch = mock.patch(
            "rhsmlib.dbus.objects.register.EntCertActionInvoker.update",
            name="update",
        )
        cls.patches["update"] = update_patch.start()
        cls.addClassCleanup(update_patch.stop)

        attach_auto_patch = mock.patch(
            "rhsmlib.dbus.objects.register.AttachService.attach_auto",
            name="attach_auto",
        )
        cls.patches["attach_auto"] = attach_auto_patch.start()
        cls.addClassCleanup(attach_auto_patch.stop)

        build_uep_patch = mock.patch(
            "rhsmlib.dbus.base_object.BaseObject.build_uep",
            name="build_uep",
        )
        cls.patches["build_uep"] = build_uep_patch.start()
        cls.addClassCleanup(build_uep_patch.stop)

        super().setUpClass()

    def setUp(self) -> None:
        self.patches["update"].return_value = None

        super().setUp()

    def test_Register(self):
        expected = json.loads(CONSUMER_CONTENT_JSON)
        self.patches["register"].return_value = expected
        self.patches["is_registered"].return_value = False

        result = self.obj.Register.__wrapped__(self.obj, "org", "username", "password", {}, {}, self.LOCALE)
        self.assertEqual(expected, json.loads(result))

    def test_Register__with_force_option(self):
        expected = json.loads(CONSUMER_CONTENT_JSON)
        self.patches["register"].return_value = expected
        self.patches["unregister"].return_value = None
        self.patches["is_registered"].return_value = True

        result = self.obj.Register.__wrapped__(
            self.obj, "org", "username", "password", {"force": True}, {}, self.LOCALE
        )
        self.assertEqual(expected, json.loads(result))

    def test_Register__already_registered(self):
        expected = json.loads(CONSUMER_CONTENT_JSON)
        self.patches["register"].return_value = expected
        self.patches["unregister"].return_value = None
        self.patches["is_registered"].return_value = True

        with self.assertRaisesRegex(dbus.DBusException, "This system is already registered."):
            self.obj.Register.__wrapped__(self.obj, "org", "username", "password", {}, {}, self.LOCALE)

    def test_Register__enable_content(self):
        """Test including 'enable_content' in entitlement mode with no content."""
        expected = json.loads(CONSUMER_CONTENT_JSON)
        self.patches["register"].return_value = expected
        self.patches["attach_auto"].return_value = []
        self.patches["is_registered"].return_value = False

        result = self.obj.Register.__wrapped__(
            self.obj, "org", "username", "password", {"enable_content": "1"}, {}, self.LOCALE
        )
        self.assertEqual(expected, json.loads(result))

    def test_Register__enable_content_with_content(self):
        """Test including 'enable_content' in entitlement mode with some content."""
        expected = json.loads(ENABLED_CONTENT)
        self.patches["register"].return_value = json.loads(CONSUMER_CONTENT_JSON)
        self.patches["attach_auto"].return_value = expected
        self.patches["is_registered"].return_value = False

        result = self.obj.Register.__wrapped__(
            self.obj, "org", "username", "password", {"enable_content": "1"}, {}, self.LOCALE
        )
        self.assertEqual(expected, json.loads(result)["enabledContent"])

    def test_Register__enable_content__sca(self):
        """Test including 'enable_content' in SCA mode."""
        expected = json.loads(CONSUMER_CONTENT_JSON_SCA)
        self.patches["register"].return_value = expected
        self.patches["is_registered"].return_value = False

        result = self.obj.Register.__wrapped__(
            self.obj, "org", "username", "password", {"enable_content": "1"}, {}, self.LOCALE
        )
        self.assertEqual(expected, json.loads(result))

    def test_GetOrgs(self):
        self.patches["is_registered"].return_value = False
        mock_cp = mock.Mock(spec=connection.UEPConnection, name="UEPConnection")
        mock_cp.username = "username"
        mock_cp.password = "password"
        mock_cp.getOwnerList = mock.Mock()
        mock_cp.getOwnerList.return_value = json.loads(OWNERS_CONTENT_JSON)
        self.patches["build_uep"].return_value = mock_cp

        expected = json.loads(OWNERS_CONTENT_JSON)
        result = self.obj.GetOrgs.__wrapped__(self.obj, "username", "password", {}, self.LOCALE)
        self.assertEqual(expected, json.loads(result))

    def test_RegisterWithActivationKeys(self):
        expected = json.loads(CONSUMER_CONTENT_JSON)
        self.patches["is_registered"].return_value = False
        self.patches["register"].return_value = expected

        result = self.obj.RegisterWithActivationKeys.__wrapped__(
            self.obj,
            "username",
            ["key1", "key2"],
            {},
            {"host": "localhost", "port": "8443", "handler": "/candlepin"},
            self.LOCALE,
        )
        self.assertEqual(expected, json.loads(result))

    def test_RegisterWithActivationKeys__already_registered(self):
        expected = json.loads(CONSUMER_CONTENT_JSON)
        self.patches["is_registered"].return_value = True
        self.patches["register"].return_value = expected

        with self.assertRaisesRegex(dbus.DBusException, "This system is already registered."):
            self.obj.RegisterWithActivationKeys.__wrapped__(
                self.obj,
                "username",
                ["key1", "key2"],
                {},
                {"host": "localhost", "port": "8443", "handler": "/candlepin"},
                self.LOCALE,
            )

    def test_RegisterWithActivationKeys__with_force_option(self):
        expected = json.loads(CONSUMER_CONTENT_JSON)
        self.patches["is_registered"].return_value = True
        self.patches["unregister"].return_value = None
        self.patches["register"].return_value = expected

        result = self.obj.RegisterWithActivationKeys.__wrapped__(
            self.obj,
            "username",
            ["key1", "key2"],
            {"force": True},
            {"host": "localhost", "port": "8443", "handler": "/candlepin"},
            self.LOCALE,
        )
        self.assertEqual(expected, json.loads(result))
